import numpy as np

from ..datalog.basic_representation import DatalogProgram
from ..datalog.expression_processing import dependency_matrix
from ..expression_pattern_matching import add_match
from ..expression_walker import ExpressionWalker
from ..expressions import FunctionApplication, Symbol
from ..logic import Implication, Union
from ..type_system import get_generic_type
from .expressions import PROB, ProbabilisticQuery
from .stratification import reachable_code_from_query


def probabilistic_postprocess_magic_rules(
    program: DatalogProgram, query: Implication, magic_rules: Union
):
    """
    Within language probabilistic queries (wlq) cannot be solved when they
    depend on one another, which is something that Magic Sets rewrite can
    introduce.

    This post processing step transforms a program where a wlq depends on
    another, i.e.

    A(x, PROB) :- B(x), C(x)
    C(x) :- D(x, p)
    D(x, PROB) :- E(x)

    into

    A(x, PROB) :- B(x), C(x)
    C(x) :- D2(x)
    D(x, PROB) :- D2(x)
    D2(x) :- E(x)
    """
    idb = list(reachable_code_from_query(query, program).formulas)
    idb_symbs, dep_mat = dependency_matrix(program, idb)
    wlq_symbs = set(program.within_language_prob_queries()).intersection(
        idb_symbs
    )
    wlq_symb_idxs = {idb_symbs.index(wlq_symb) for wlq_symb in wlq_symbs}
    replaced = set()
    wlq_exprs = set()
    rules_to_update = set()
    for wlq_symb_idx in wlq_symb_idxs:
        wlq_symb_deps = _wlq_dependencies(wlq_symb_idx, dep_mat, wlq_symb_idxs)
        for to_replace in wlq_symb_deps:
            wlq_symb = idb_symbs[to_replace]
            if wlq_symb not in replaced:
                wlq_exprs.add(
                    program.within_language_prob_queries()[wlq_symb].consequent
                )
                code = reachable_code_from_query(
                    program.within_language_prob_queries()[wlq_symb],
                    program,
                )
                for rule in code.formulas:
                    try:
                        rules_to_update.add(magic_rules.formulas.index(rule))
                    except ValueError:
                        pass

    processed_idb = update_rules_with_new_prob_expressions(
        magic_rules,
        rules_to_update,
        wlq_exprs,
    )
    query_predicate = query.consequent
    processed_query = next(
        r
        for r in processed_idb
        if r.consequent.functor == query_predicate.functor
    )
    return processed_query, Union(processed_idb)


def update_rules_with_new_prob_expressions(
    magic_rules, rules_to_update, wlq_exprs
):
    processed = []
    walker = ReplaceWLQWalker(wlq_exprs)
    for idx, rule in enumerate(magic_rules.formulas):
        if idx in rules_to_update:
            updated = walker.walk(rule)
            if isinstance(updated, tuple):
                processed.extend(updated)
            else:
                processed.append(updated)
        else:
            processed.append(rule)
    return processed


def _wlq_dependencies(wlq_symb_idx, dep_mat, wlq_symb_idxs):
    stack = [wlq_symb_idx]
    deps = set()
    seen = set()
    while stack:
        dep_idxs = set(
            np.argwhere(dep_mat[stack.pop()].astype(bool)).flatten()
        )
        deps.update(wlq_symb_idxs.intersection(dep_idxs))
        stack += list(dep_idxs - seen)
        seen.update(dep_idxs)
    return deps


def _is_prob_arg(arg):
    return isinstance(arg, ProbabilisticQuery) or (
        get_generic_type(type(arg)) is FunctionApplication
        and arg.functor == PROB
    )


class ReplaceWLQWalker(ExpressionWalker):
    """
    Initialize the walker with matchings {S{D} : S{D2}}.
    The walker will then match rules that look like
    D(x, y, PROB) :- A(x, y)

    and replace them by

    D(x, y, PROB) :- D2(x, y)
    D2(x, y) :- A(x, y)
    """

    def __init__(self, wlq_expressions):
        self.wlq_expressions = wlq_expressions
        self.symbols = {expr.functor for expr in self.wlq_expressions}

    @add_match(
        Implication,
        lambda implication: isinstance(
            implication.consequent, FunctionApplication
        )
        and any(_is_prob_arg(arg) for arg in implication.consequent.args),
    )
    def replace_probabilistic_queries(self, impl):
        functor = impl.consequent.functor
        if functor in self.symbols:
            new_symbol = Symbol(f"{functor.name}_noprob")
            non_prob_args = tuple(
                arg for arg in impl.consequent.args if not _is_prob_arg(arg)
            )
            new_wlq = Implication(
                impl.consequent.functor(*impl.consequent.args),
                new_symbol(*non_prob_args),
            )
            new_non_wlq = Implication(
                new_symbol(*non_prob_args), impl.antecedent
            )
            return (new_non_wlq, new_wlq)
        return impl

    @add_match(Implication)
    def implication(self, impl):
        return Implication(impl.consequent, self.walk(impl.antecedent))

    @add_match(FunctionApplication)
    def replace_probabilistic_predicate(self, expr):
        if expr.functor in self.symbols:
            wlq_expr = next(
                e for e in self.wlq_expressions if e.functor == expr.functor
            )
            non_prob_args = tuple(
                arg
                for arg, wlq_arg in zip(expr.args, wlq_expr.args)
                if not _is_prob_arg(wlq_arg)
            )
            new_expr = Symbol(f"{expr.functor.name}_noprob")(*non_prob_args)
            return new_expr
        return self.process_expression(expr)



import logging
import time
import numpy as np
import pandas as pd
import dask.dataframe as dd
from functools import reduce
from sqlalchemy import table, column, and_, select
from sqlalchemy.dialects import postgresql
from dask_sql import Context
from neurolang.utils.relational_algebra_set import (
    pandas,
    dask_sql,
)

LOGGING_CONFIG = {
    "version": 1,
    "formatters": {
        "standard": {
            "format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"
        },
    },
    "handlers": {
        "default": {
            "level": "DEBUG",
            "formatter": "standard",
            "class": "logging.StreamHandler",
        },
    },
    "loggers": {
        "": {  # root logger
            "handlers": ["default"],
            "level": "WARNING",
            "propagate": False,
        },
        "neurolang.utils.relational_algebra_set": {
            "handlers": ["default"],
            "level": "INFO",
            "propagate": False,
        },
        "dask_sql.context": {
            "handlers": ["default"],
            "level": "DEBUG",
            "propagate": False,
        },
    },
}
logging.config.dictConfig(LOGGING_CONFIG)


class TimeLeftNaturalJoins:
    params = [
        [10 ** 4, 10 ** 5],
        [10],
        [3],
        [6, 12],
        [0.75],
        [pandas, dask_sql],
    ]

    param_names = [
        "rows",
        "cols",
        "number of join columns",
        "number of chained joins",
        "ratio of dictinct elements",
        "RAS module to test",
    ]

    def setup(self, N, ncols, njoin_columns, njoins, distinct_r, module):
        dfs = _generate_dataframes(N, ncols, njoin_columns, njoins, distinct_r)
        self.sets = [
            module.NamedRelationalAlgebraFrozenSet(df.columns, df)
            for df in dfs
        ]

    def time_ra_left_naturaljoin(
        self, N, ncols, njoin_columns, njoins, distinct_r, module
    ):
        res = reduce(lambda a, b: a.left_naturaljoin(b), self.sets)
        post_process_result(self.sets, res)


class TimeChainedNaturalJoins:
    params = [
        [10 ** 4, 10 ** 5],
        [10],
        [3],
        [6],
        [0.75],
        [pandas, dask_sql],
    ]

    param_names = [
        "rows",
        "cols",
        "number of join columns",
        "number of chained joins",
        "ratio of dictinct elements",
        "RAS module to test",
    ]

    timeout = 60 * 3

    def setup(self, N, ncols, njoin_columns, njoins, distinct_r, module):
        dfs = _generate_dataframes(N, ncols, njoin_columns, njoins, distinct_r)
        self.sets = [
            module.NamedRelationalAlgebraFrozenSet(df.columns, df)
            for df in dfs
        ]

    def time_ra_naturaljoin_hard(
        self, N, ncols, njoin_columns, njoins, distinct_r, module
    ):
        res = reduce(lambda a, b: a.naturaljoin(b), self.sets)
        post_process_result(self.sets, res)

    def time_ra_naturaljoin_easy(
        self, N, ncols, njoin_columns, njoins, distinct_r, module
    ):
        res = reduce(lambda a, b: a.naturaljoin(b), self.sets[::-1])
        post_process_result(self.sets, res)


class TimeRawMerge:
    params = [[10 ** 4, 10 ** 5], [10], [3], [6, 12], [0.75], [dd, pd]]

    param_names = [
        "rows",
        "cols",
        "number of join columns",
        "number of chained joins",
        "ratio of dictinct elements",
        "lib",
    ]

    def setup(self, N, ncols, njoin_columns, njoins, distinct_r, lib):
        self.dfs = _generate_dataframes(
            N, ncols, njoin_columns, njoins, distinct_r
        )
        if lib == dd:
            self.dfs = [dd.from_pandas(d, npartitions=1) for d in self.dfs]

    def time_merge(self, N, ncols, njoin_columns, njoins, distinct_r, lib):
        on = list(self.dfs[0].columns[:njoin_columns])
        res = reduce(lambda a, b: lib.merge(a, b, on=on), self.dfs)
        if lib == dd:
            df = res.compute()


class TimeDaskSQLJoins:
    params = [[10 ** 4, 10 ** 5], [10], [3], [6], [0.75]]

    param_names = [
        "rows",
        "cols",
        "number of join columns",
        "number of chained joins",
        "ratio of dictinct elements",
    ]

    def setup(self, N, ncols, njoin_columns, njoins, distinct_r):
        self.dfs = _generate_dataframes(
            N, ncols, njoin_columns, njoins, distinct_r
        )
        self.dfs = [dd.from_pandas(d, npartitions=1) for d in self.dfs]
        self.join_cols = [
            c for c in self.dfs[0].columns if c in self.dfs[1].columns
        ]
        self.ctx = Context()
        self._create_tables()
        self._create_sql_query()

    def _create_tables(self):
        self.tables = []
        for i, df in enumerate(self.dfs):
            _table_name = f"table_{i:03}"
            self.ctx.create_table(_table_name, df)
            _table = table(_table_name, *[column(c) for c in df.columns])
            self.tables.append(_table)

    def _create_sql_query(self):
        left = self.tables[0]
        joinq = left
        select_cols = list(left.c)
        for right in self.tables[1:]:
            on = and_(
                left.c.get(col) == right.c.get(col) for col in self.join_cols
            )
            joinq = joinq.join(right, on)
            select_cols += [c for c in right.c if c.name not in self.join_cols]
        query = select(*select_cols).select_from(joinq)
        self.sql_query = str(
            query.compile(
                dialect=postgresql.dialect(),
                compile_kwargs={"literal_binds": True},
            )
        )

    def time_joins(self, N, ncols, njoin_columns, njoins, distinct_r):
        start = time.perf_counter()
        print(f"Processing SQL query: {self.sql_query}")
        res = self.ctx.sql(self.sql_query)
        stop = time.perf_counter()
        print(f"Processing SQL query took {stop-start:0.4f} s.")
        start = time.perf_counter()
        print("Computing dask dataframe")
        res.compute()
        stop = time.perf_counter()
        print(f"Computing dask dataframe took {stop-start:0.4f} s.")
        # Visualize task graph
        # res.visualize('taskgraph.png')
        return res


class TimeEquiJoin:
    params = [
        [10 ** 4, 10 ** 5],
        [10],
        [3],
        [6, 12],
        [0.75],
        [pandas, dask_sql],
    ]

    param_names = [
        "rows",
        "cols",
        "number of join columns",
        "number of chained joins",
        "ratio of dictinct elements",
        "RAS module to test",
    ]

    def setup(self, N, ncols, njoin_columns, njoins, distinct_r, module):
        dfs = _generate_dataframes(N, ncols, njoin_columns, njoins, distinct_r)
        for d in dfs:
            d.columns = pd.RangeIndex(ncols)
        self.sets = [module.RelationalAlgebraFrozenSet(df) for df in dfs]

    def time_ra_equijoin(
        self, N, ncols, njoin_columns, njoins, distinct_r, module
    ):
        res = reduce(
            lambda a, b: a.equijoin(b, [(i, i) for i in range(njoin_columns)]),
            self.sets,
        )

        post_process_result(self.sets, res)


def post_process_result(sets, result):
    if isinstance(result, dask_sql.RelationalAlgebraFrozenSet):
        # print(result)
        # Fetch one seems slower than _fetchall. Need to investigate.
        result._fetchall()
        # result.fetch_one()


def _generate_dataframes(N, ncols, njoin_columns, njoins, distinct_r):
    """
    Generate njoins dataframes of decreasing size. The first dataframe has
    shape N x ncols, while each dataframe after that has shape 
    (N / (i + 1) x ncols).
    The first njoin_columns cols of each dataframe have identical rows to
    perform joins on.
    """
    join_columns = [hex(x) for x in range(njoin_columns)]
    rstate = np.random.RandomState(0)
    keys = pd.DataFrame(
        rstate.randint(0, N * distinct_r, size=(N, njoin_columns)),
        columns=join_columns,
    )
    sets = []
    for i in range(njoins):
        # Take a sample of the default keys.
        skeys = keys.sample(frac=1 / (i + 1), random_state=rstate)
        skeys = pd.DataFrame(
            np.tile(skeys.to_numpy(), (njoins - i, 1)), columns=join_columns,
        )
        # Generate random data for the rest of the set
        cols = [hex(ncols * i + x) for x in range(njoin_columns, ncols)]
        df = pd.concat(
            [
                skeys,
                pd.DataFrame(
                    rstate.randint(
                        0,
                        N * distinct_r,
                        size=(skeys.shape[0], ncols - njoin_columns),
                    ),
                    columns=cols,
                ),
            ],
            axis=1,
        )
        sets.append(df)
    return sets



"""
Implentation of probabilistic query resolution for
hierarchical queries [^1]. Using this we apply the small dichotomy
theorem [^1, ^2]:
Let Q be a conjunctive query without self-joins or a non-repeating relational
algebra expression. Then:
* If Q is hierarchical, then P(Q) is in polynomial time, and can be computed
using only the lifted inference rules for join, negation, union, and
existential quantifier.

* If Q is not hierarchical, then P(Q) is #P-hard in the size of the database.

[^1]: Robert Fink and Dan Olteanu. A dichotomy for non-repeating queries with
negation in probabilistic databases. In Proceedings of the 33rd ACM
SIGMOD-SIGACT-SIGART Symposium on Principles of Database Systems, PODS ’14,
pages 144–155, New York, NY, USA, 2014. ACM.

[^2]: Nilesh N. Dalvi and Dan Suciu. Efficient query evaluation on
probabilistic databases. VLDB J., 16(4):523–544, 2007.
"""

import logging
from collections import defaultdict
from typing import AbstractSet

from ..datalog.expression_processing import (
    EQ,
    UnifyVariableEqualities,
    extract_logic_atoms,
    extract_logic_free_variables,
    extract_logic_predicates,
    flatten_query,
    remove_conjunction_duplicates
)
from ..datalog.translate_to_named_ra import TranslateToNamedRA
from ..exceptions import UnsupportedSolverError
from ..expression_walker import ReplaceExpressionWalker
from ..expressions import Constant, Symbol
from ..logic import FALSE, Conjunction, Disjunction, Implication, Negation
from ..logic.transformations import (
    GuaranteeConjunction,
    MakeExistentialsImplicit,
    RemoveTrivialOperations,
    convert_to_pnf_with_dnf_matrix
)
from ..relational_algebra import (
    NamedRelationalAlgebraFrozenSet,
    Projection,
    str2columnstr_constant
)
from ..relational_algebra_provenance import ProvenanceAlgebraSet
from ..utils import log_performance
from ..utils.orderedset import OrderedSet
from .exceptions import NotHierarchicalQueryException
from .expression_processing import lift_optimization_for_choice_predicates
from .probabilistic_ra_utils import (
    generate_probabilistic_symbol_table_for_query,
    is_atom_a_probabilistic_choice_relation,
    is_atom_a_probabilistic_fact_relation
)
from .query_resolution import (
    generate_provenance_query_solver,
    lift_solve_marg_query,
    compute_projections_needed_to_reintroduce_head_terms
)
from .shattering import shatter_constants

LOG = logging.getLogger(__name__)


def is_hierarchical_without_self_joins(query):
    """
    Let Q be first-order formula. For each variable x denote at(x) the
    set of atoms that contain the variable x. We say that Q is hierarchical
    if forall x, y one of the following holds:
    at(x) ⊆ at(y) or at(x) ⊇ at(y) or at(x) ∩ at(y) = ∅.
    """
    query = convert_to_pnf_with_dnf_matrix(query)
    antecedent = MakeExistentialsImplicit().walk(query)
    if isinstance(antecedent, Disjunction):
        return False

    has_self_joins, atom_set = extract_atom_sets_and_detect_self_joins(query)
    has_negative_atoms = any(
        isinstance(predicate, Negation)
        for predicate in extract_logic_predicates(query)
    )

    if has_self_joins or has_negative_atoms:
        return False

    variables = list(atom_set)
    for i, v in enumerate(variables):
        at_v = atom_set[v]
        for v2 in variables[i + 1:]:
            at_v2 = atom_set[v2]
            if not (at_v <= at_v2 or at_v2 <= at_v or at_v.isdisjoint(at_v2)):
                LOG.info(
                    "Not hierarchical on variables %s %s", v.name, v2.name
                )
                return False

    return True


def extract_atom_sets_and_detect_self_joins(query):
    has_self_joins = False
    predicates = extract_logic_atoms(query)
    predicates = set(pred for pred in predicates if not pred.functor == EQ)
    seen_predicate_functor = set()
    atom_set = defaultdict(set)
    for predicate in predicates:
        functor = predicate.functor
        if functor in seen_predicate_functor:
            LOG.info("Not hierarchical self join on variables %s", functor)
            has_self_joins = True
        seen_predicate_functor.add(functor)
        for variable in predicate.args:
            if not isinstance(variable, Symbol):
                continue
            atom_set[variable].add(functor)
    return has_self_joins, atom_set


def solve_succ_query(query, cpl_program, run_relational_algebra_solver=True):
    """
    Solve a SUCC query on a CP-Logic program.

    Parameters
    ----------
    query : Implication
        SUCC query of the form `ans(x) :- P(x)`.
    cpl_program : CPLogicProgram
        CP-Logic program on which the query should be solved.
    run_relational_algebra_solver: bool, default True
        When true the result's `relation` attribute is
        a NamedRelationalAlgebraFrozenSet,
        when false the attribute is the relational algebra expression that
        produces the such set.

    Returns
    -------
    ProvenanceAlgebraSet
        Provenance set labelled with probabilities for each tuple in the result
        set.

    """

    with log_performance(
        LOG,
        "Preparing query %s",
        init_args=(query.consequent.functor.name,),
    ):
        flat_query_body = flatten_query(query.antecedent, cpl_program)

    if flat_query_body == FALSE or (
        isinstance(flat_query_body, Conjunction)
        and any(conjunct == FALSE for conjunct in flat_query_body.formulas)
    ):
        head_var_names = tuple(
            term.name for term in query.consequent.args
            if isinstance(term, Symbol)
        )
        return ProvenanceAlgebraSet(
            Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
                ("_p_",) + head_var_names)
            ),
            str2columnstr_constant("_p_"),
        )

    if isinstance(flat_query_body, Conjunction):
        flat_query_body = remove_conjunction_duplicates(flat_query_body)

    with log_performance(LOG, "Translation and lifted optimisation"):
        flat_query_body = GuaranteeConjunction().walk(
            lift_optimization_for_choice_predicates(
                flat_query_body, cpl_program
            )
        )
        flat_query = Implication(query.consequent, flat_query_body)

        symbol_table = generate_probabilistic_symbol_table_for_query(
            cpl_program, flat_query_body
        )
        unified_query = UnifyVariableEqualities().walk(flat_query)
        _verify_fact_and_choice_e_quantification(symbol_table, unified_query)
        original_symbols = set(symbol_table)
        shattered_query_antecedent = shatter_constants(
            unified_query.antecedent, symbol_table
        )
        shattered_symbols = set(symbol_table) - original_symbols

        shattered_query_antecedent = convert_to_pnf_with_dnf_matrix(
            shattered_query_antecedent
        )
        shattered_query_antecedent = (
            RemoveTrivialOperations()
            .walk(MakeExistentialsImplicit().walk(shattered_query_antecedent))
        )
        if isinstance(shattered_query_antecedent, Disjunction):
            raise UnsupportedSolverError("Query not conjunctive")

        shattered_query = Implication(
            unified_query.consequent,
            shattered_query_antecedent
        )
        probabilistic_predicates = \
            _extract_antecedent_probabilistic_predicates(
                shattered_query,
                symbol_table
            )
        shattered_query_probabilistic_body = Conjunction(
            tuple(probabilistic_predicates)
        )
        if not is_hierarchical_without_self_joins(
            shattered_query_probabilistic_body
        ):
            LOG.info(
                "Query with conjunctions %s not hierarchical",
                shattered_query_probabilistic_body.formulas,
            )
            raise NotHierarchicalQueryException(
                "Query not hierarchical, algorithm can't be applied"
            )

        ra_query = TranslateToNamedRA().walk(shattered_query.antecedent)
        replace_shattered_symbols = ReplaceExpressionWalker(
            {k: symbol_table[k] for k in shattered_symbols}
        )
        ra_query = replace_shattered_symbols.walk(ra_query)
        # project on query's head variables
        ra_query = _project_on_query_head(ra_query, shattered_query)
        # re-introduce head variables potentially removed by unification
        needed_projections = compute_projections_needed_to_reintroduce_head_terms(
            ra_query, flat_query, unified_query
        )

    query_solver = generate_provenance_query_solver(
        symbol_table, run_relational_algebra_solver,
        needed_projections=needed_projections
    )

    with log_performance(LOG, "Run RAP query %s", init_args=(ra_query,)):
        prob_set_result = query_solver.walk(ra_query)

    return prob_set_result


def _verify_fact_and_choice_e_quantification(symbol_table, shattered_query):
    e_vars = (
        extract_logic_free_variables(shattered_query.antecedent) -
        extract_logic_free_variables(shattered_query.consequent)
    )
    e_vars_choice = {
        var for var in e_vars
        if any(
            var in set(atom.args)
            for atom in extract_logic_atoms(shattered_query.antecedent)
            if is_atom_a_probabilistic_choice_relation(atom, symbol_table)
        )
    }
    if any(
        (e_vars & atom.args) and
        is_atom_a_probabilistic_fact_relation(atom, symbol_table) and
        (e_vars & atom.args) > e_vars_choice
        for atom in extract_logic_atoms(shattered_query.antecedent)
    ):
        raise UnsupportedSolverError(
                "Cannot solve queries with existentially quantified variable "
                "occurring in probabilistic fact but not in any probabilistic "
                "choice, or disjunctive queries"
            )


def _extract_antecedent_probabilistic_predicates(datalog_rule, symbol_table):
    probabilistic_predicates = []
    for predicate in extract_logic_predicates(datalog_rule):
        atom = predicate
        while isinstance(atom, Negation):
            atom = atom.formula
        if isinstance(atom.functor, Symbol) and (
            is_atom_a_probabilistic_choice_relation(atom, symbol_table)
            or
            is_atom_a_probabilistic_fact_relation(atom, symbol_table)
        ):
            probabilistic_predicates.append(predicate)
    return probabilistic_predicates


def _project_on_query_head(provset, query):
    proj_cols = tuple(
        OrderedSet(
            str2columnstr_constant(arg.name)
            for arg in query.consequent.args
            if isinstance(arg, Symbol)
        )
    )
    return Projection(provset, proj_cols)


def solve_marg_query(rule, cpl):
    """
    Solve a MARG query on a CP-Logic program.
    Parameters
    ----------
    query : Implication
        Consequent must be of type `Condition`.
        MARG query of the form `ans(x) :- P(x)`.
    cpl_program : CPLogicProgram
        CP-Logic program on which the query should be solved.
    Returns
    -------
    ProvenanceAlgebraSet
        Provenance set labelled with probabilities for each tuple in the result
        set.
    """
    return lift_solve_marg_query(rule, cpl, solve_succ_query)



from collections import defaultdict
from typing import AbstractSet, Any, Mapping

from ..datalog.wrapped_collections import (
    WrappedNamedRelationalAlgebraFrozenSet
)
from ..expression_walker import PatternWalker, add_match
from ..expressions import Constant, Expression, Symbol
from ..relational_algebra import (
    ColumnInt,
    RelationalAlgebraOperation,
    UnaryRelationalAlgebraOperation
)


class ProbabilisticFactSet(RelationalAlgebraOperation):
    def __init__(self, relation, probability_column):
        self.relation = relation
        self.probability_column = probability_column


class ProbabilisticChoiceSet(RelationalAlgebraOperation):
    def __init__(self, relation, probability_column):
        self.relation = relation
        self.probability_column = probability_column


class DeterministicFactSet(RelationalAlgebraOperation):
    def __init__(self, relation):
        self.relation = relation


class NonLiftable(RelationalAlgebraOperation):
    """Represents a logic-based probabilistic query
    which could not be lifted into a safe plan.
    """
    def __init__(self, non_liftable_query):
        self.non_liftable_query = non_liftable_query

    def __repr__(self):
        return (
            "NonLiftable"
            f"({self.non_liftable_query})"
        )


def generate_probabilistic_symbol_table_for_query(
    cpl_program, query_predicate
):
    """Generate a symbol table adding a fresh symbol for
    each probabilistic/deterministic table and wrapping
    it into one of the classes
    DeterministicFactSet, ProbabilisticFactSet, ProbabilisticChoiceSet

    Parameters
    ----------
    cpl_program : neurolang.probabilistic.cplogic.CPLogicProgram
        CPLogic program representation indexing the query's symbols.
    query_predicate : neurolang.logic.LogicOperator
        Logic expression representing a query from where
        to extract the predicates which are represented as Symbols.

    Returns
    -------
    dict or Mapping
        symbol table with relationship symbols of query_predicate.
        The tables has neurolang.expressions.Symbol as keys and
        relational algebra representations as values
    """
    symbol_table = defaultdict(
        lambda: DeterministicFactSet(
            Constant[AbstractSet](
                WrappedNamedRelationalAlgebraFrozenSet(
                    columns=tuple()
                )
            )
        )
    )
    classify_and_wrap_symbols(
        cpl_program.probabilistic_facts(), query_predicate,
        symbol_table, ProbabilisticFactSet
    )

    classify_and_wrap_symbols(
        cpl_program.probabilistic_choices(), query_predicate,
        symbol_table, ProbabilisticChoiceSet
    )

    classify_and_wrap_symbols(
        cpl_program.extensional_database(), query_predicate,
        symbol_table, lambda s, _: DeterministicFactSet(s)
    )

    return symbol_table


def classify_and_wrap_symbols(
    ra_set_dict, query_predicate, symbol_table, wrapper
):
    for predicate_symbol, facts in ra_set_dict.items():
        if predicate_symbol not in query_predicate._symbols:
            continue

        fresh_symbol = Symbol.fresh()
        symbol_table[predicate_symbol] = wrapper(
            fresh_symbol,
            Constant[ColumnInt](ColumnInt(0))
        )
        symbol_table[fresh_symbol] = facts


class GetProbabilisticSetAtom(PatternWalker):
    def __init__(self, symbol_table):
        self.symbol_table = symbol_table

    @add_match(
        UnaryRelationalAlgebraOperation,
        lambda expression: not isinstance(
            expression,
            (
                DeterministicFactSet,
                ProbabilisticFactSet,
                ProbabilisticChoiceSet
            )
        )
    )
    def projection(self, expression):
        return self.walk(expression.relation)

    @add_match(Symbol)
    def resolve_symbol(self, expression):
        if expression in self.symbol_table:
            return self.walk(self.symbol_table[expression])
        else:
            return expression

    @add_match(...)
    def default(self, expression):
        return expression


def is_atom_a_deterministic_relation(
    atom: Expression, symbol_table: Mapping[Any, Expression]
) -> bool:
    """Returns if a particular expression is a deterministic relation

    Parameters
    ----------
    atom : Expression
        neurolang Expression
    symbol_table : Mapping[Any, Expression]
        mapping maching symbols to expressions.

    Returns
    -------
    bool
        True if the Expression is a relational algebra expression
        containing a single deterministic relation.
    """
    gpsa = GetProbabilisticSetAtom(symbol_table)
    relation = gpsa.walk(atom.functor)
    return (
        isinstance(relation, DeterministicFactSet)
    )


def is_atom_a_probabilistic_fact_relation(
    atom: Expression, symbol_table: Mapping[Any, Expression]
) -> bool:
    """Returns if a particular expression is a tuple independent
    fact set relation

    Parameters
    ----------
    atom : Expression
        neurolang Expression
    symbol_table : Mapping[Any, Expression]
        mapping maching symbols to expressions.

    Returns
    -------
    bool
        True if the Expression is a relational algebra expression
        containing a single independent fact set relation.
    """
    gpsa = GetProbabilisticSetAtom(symbol_table)
    relation = gpsa.walk(atom.functor)
    return (
        isinstance(relation, ProbabilisticFactSet)
    )


def is_atom_a_probabilistic_choice_relation(
    atom: Expression, symbol_table: Mapping[Any, Expression]
) -> bool:
    """Returns if a particular expression is a choice
    relation

    Parameters
    ----------
    atom : Expression
        neurolang Expression
    symbol_table : Mapping[Any, Expression]
        mapping maching symbols to expressions.

    Returns
    -------
    bool
        True if the Expression is a relational algebra expression
        containing a single choice relation.
    """
    gpsa = GetProbabilisticSetAtom(symbol_table)
    relation = gpsa.walk(atom.functor)
    return (
        isinstance(relation, ProbabilisticChoiceSet)
    )



from collections import namedtuple
from pathlib import Path
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from ..frontend.probabilistic_frontend import NeurolangPDL

from ..datalog.basic_representation import DatalogProgram
from ..exceptions import InvalidCommandExpression, UnsupportedProgramError
from ..expression_walker import ExpressionBasicEvaluator
from ..expressions import Command, Constant, Symbol
from ..probabilistic.cplogic.program import CPLogicMixin
from ..commands import CommandsMixin


class Datalog(
    CommandsMixin, DatalogProgram, ExpressionBasicEvaluator, CPLogicMixin,
):
    pass


def test_invalid_commands_raise_exceptions():
    cmd = Command(
        Symbol("load_csv"), (Constant("https://somerandomurl.csv"),), ()
    )

    datalog = Datalog()
    with pytest.raises(InvalidCommandExpression):
        datalog.walk(cmd)

    cmd = Command(Symbol("print"), (Symbol("A"),), ())
    with pytest.raises(UnsupportedProgramError):
        datalog.walk(cmd)


@patch("neurolang.commands.pd.read_csv")
def test_load_csv_command_adds_tuple_set(mock_pd_readcsv):
    mock_study_ids = pd.DataFrame({"A": [1, 2], "B": [3, 4]})
    mock_pd_readcsv.return_value = mock_study_ids

    url = "https://github.com/neuroquery/.../data-neuroquery_version-1_metadata.tsv.gz"

    studies = Symbol("Studies")
    cmd = Command(
        Symbol("load_csv"),
        (studies, Constant(url),),
        ((Symbol("sep"), Constant("\t")),),
    )

    datalog = Datalog()
    datalog.walk(cmd)

    mock_pd_readcsv.assert_called_with(url, sep="\t")
    assert studies in datalog.symbol_table
    assert (
        datalog.symbol_table[studies].value
        == mock_study_ids.to_records(index=False).tolist()
    )


@patch("neurolang.commands.nibabel.load")
@patch("neurolang.commands.pd.read_csv")
@patch("neurolang.commands._fetch_files")
def test_load_atlas_command(mock_fetch_files, mock_read_csv, mock_nib_load):
    """
    Test that load_atlas command can load destrieux atlas in neurolang
    """
    # setup mocks for fetch methods
    mock_fetch_files.return_value = [
        "/home/users/mock_atlas_file",
        "/home/users/mock_labels_file",
    ]
    mock_labels = pd.DataFrame(
        {"id": [1, 2], "labels": ["L S_central", "L S_postcentral"]}
    )
    mock_read_csv.return_value = mock_labels
    shape = (3, 3, 3)
    dataobj = np.arange(27).reshape(shape)
    affine = np.eye(4)
    mock_image = namedtuple("SpatialImage", ["shape", "affine", "dataobj"])(
        shape, affine, dataobj
    )
    mock_nib_load.return_value = mock_image

    # Create command and walk it
    url = "https://www.nitrc.org/frs/download.php/11942/destrieux2009.tgz"
    labels_url = "destrieux2009_rois_labels_lateralized.csv"
    atlas_url = "destrieux2009_rois_lateralized.nii.gz"
    destrieux = Symbol("Destrieux")
    cmd = Command(
        Symbol("load_atlas"),
        (destrieux, Constant(atlas_url), Constant(labels_url), Constant(url),),
        (),
    )

    datalog = Datalog()
    datalog.walk(cmd)

    mock_fetch_files.assert_called_with(
        Path.home() / "neurolang_data",
        [
            (atlas_url, url, {"uncompress": True}),
            (labels_url, url, {"uncompress": True}),
        ],
    )
    assert destrieux in datalog.symbol_table
    assert len(datalog.symbol_table[destrieux].value) == 2


def test_command_end2end():
    """
    Create a PDL engine and use the .load_atlas command to load the data.
    """
    nl = NeurolangPDL()

    @nl.add_symbol
    def startswith(prefix: str, s: str) -> bool:
        return s.startswith(prefix)

    query = """.load_atlas(Destrieux, "destrieux2009_rois_lateralized.nii.gz", "destrieux2009_rois_labels_lateralized.csv", "https://www.nitrc.org/frs/download.php/11942/destrieux2009.tgz")
    LeftSulcus(name_, region) :- Destrieux(name_, region) , startswith("L S", name_)"""

    with nl.scope:
        nl.execute_datalog_program(query)
        res = nl.solve_all()

    assert len(res["LeftSulcus"]) == 31
    assert "LeftSulcus" not in nl.symbol_table



import lark

from .ast import ASTNode



class LarkASTConverter(object):

    @staticmethod
    def _default(ast):
        if isinstance(ast, lark.tree.Tree):
            return ASTNode(
                ast['parseinfo'].rule,
                ast
            )
        else:
            return ast



class NeuroLangException(Exception):
    """Base class for NeuroLang Exceptions"""

    pass


class UnexpectedExpressionError(NeuroLangException):
    pass


class NeuroLangNotImplementedError(NeuroLangException):
    pass


class ForbiddenExpressionError(NeuroLangException):
    """
    Generic exception specifying an error in the program.
    """
    pass


class ForbiddenDisjunctionError(ForbiddenExpressionError):
    """
    Probabilistic queries do not support disjunctions.

    A probabilistic choice can be added for a predicate symbol by using the
    `add_probabilistic_choice_from_tuples` method of a probabilistic Neurolang
    engine. But you cannot add multiple probabilistic facts or rule for the
    same
    predicate.

    Examples
    ----------
    ProbActivation(r, PROB(r)) :- RegionReported(r, s) & SelectedStudy(s)
    ProbActivation(r, PROB(r)) :- ~RegionReported(r, s) & SelectedStudy(s)

    This example program adds a disjunction of probabilistic queries which is
    not allowed.
    """

    pass


class ForbiddenExistentialError(ForbiddenExpressionError):
    pass


class RelationalAlgebraError(NeuroLangException):
    """
    Base class for Relational Algebra provenance exceptions.
    """

    pass


class NotConjunctiveExpression(NeuroLangException):
    """
    This expression is not conjunctive. In this case, an expression is
    conjunctive if it is a conjunction of
      - Constant
      - A function or predicate of constants
    """

    pass


class NotConjunctiveExpressionNegation(NotConjunctiveExpression):
    """
    This expression is not conjunctive. In this case, an expression is
    conjunctive if it is a conjunction of
      - Constant
      - A function or predicate of conjunctive arguments
      - A negated predicate of conjunctive arguments
    """

    pass


class NotConjunctiveExpressionNestedPredicates(NotConjunctiveExpression):
    """
    This expression is not conjunctive. In this case, an expression is
    conjunctive if it is a conjunction of
      - Constant
      - A function or predicate of conjunctive arguments
      - A quantifier of conjunctive arguments

    Note that in this case, negated predicates are not valid (negation and
    aggregation cannot be used in the same rule).

    Examples
    --------
    StudyMatchingRegionSegregationQuery(count(s), r) :-
        RegionReported(r, s) & ~RegionReported(r2, s)
        & RegionLabel(r2) & (r2 != r)

    The above expression is not conjunctive since it uses an aggregate
    function `count` in combination with a negated predicate
    `~RegionReported`.
    """

    pass


class ProjectionOverMissingColumnsError(RelationalAlgebraError):
    """
    One of the predicates in the program has wrong arguments.
    See `WrongArgumentsInPredicateError`
    """

    pass


class RelationalAlgebraNotImplementedError(
    RelationalAlgebraError, NotImplementedError
):
    """
    Neurolang was unable to match one of the relational algebra operations
    defined in the program. This is probably due to a malformed query.
    """

    pass


class ForbiddenBuiltinError(ForbiddenExpressionError):
    pass


class NotInFONegE(ForbiddenExpressionError):
    pass


class NeuroLangFrontendException(NeuroLangException):
    pass


class SymbolNotFoundError(NeuroLangException):
    """
    A symbol is being used in a rule without having been previously
    defined.
    """

    pass


class RuleNotFoundError(NeuroLangException):
    pass


class UnsupportedProgramError(NeuroLangException):
    """
    Some parts of the datalog program are (currently) unsupported.
    """
    pass


class UnsupportedQueryError(NeuroLangException):
    """
    Queries on probabilistic predicates are unsupported.
    """
    pass


class UnsupportedSolverError(NeuroLangException):
    pass


class ProtectedKeywordError(NeuroLangException):
    """
    One of the predicates in the program uses a reserved keyword.
    Reserved keywords include : {PROB, with, exists}
    """
    pass


class ForbiddenRecursivityError(UnsupportedProgramError):
    """
    The given program cannot be stratified due to recursivity.

    When using probabilistic queries, a query can
    be solved through stratification if the probabilistic and deterministic
    parts are well separated. In case there exists one within-language
    probabilistic query dependency, no probabilistic predicate should appear
    in the stratum that depends on the query.
    The same holds for aggregate or negated queries. If a rule contains an
    aggregate or negated term, all the predicates in the body of the rule
    must be computed in a previous stratum.

    Examples
    --------
    B(x) :- A(x), C(x)
    A(x) :- B(x)

    This program cannot be stratified because it contains a loop in
    the dependencies of each rule. Rule `B(x) :- A(x), C(x)` depends
    on the second rule through its occurence of the predicate `A(x)`.
    But rule `A(x) :- B(x)` in turn depends on the first rule through
    the `B(x)` predicate.
    """

    pass


class ForbiddenUnstratifiedAggregation(UnsupportedProgramError):
    """
    The given Datalog program is not valid for aggregation. Support for
    aggregation is done according to section 2.4.1 of [1]_.

    A program is valid for aggregation if it can be stratified into
    strata P1, . . . , Pn such that, if A :- ...,B,... is a rule in P such
    that A contains an aggregate term, and A is in stratum Pi while B is in
    stratum Pj, **then i > j**.

    In other terms, all the predicates in the body of a rule containing an
    aggregate function must be computed in a previous stratum. Recursion
    through aggregation is therefore not allowed in the same stratum.

    Examples
    --------
    The following datalog program is invalid for stratified aggregation
    p(X) :- q(X).
    p(sum<X>) :- p(X).


    .. [1] T. J. Green, S. S. Huang, B. T. Loo, W. Zhou,
       Datalog and Recursive Query Processing.
       FNT in Databases. 5, 105–195 (2012).
    """

    pass


class WrongArgumentsInPredicateError(NeuroLangException):
    """
    One of the predicates in the query has the wrong number of arguments.

    Examples
    --------
    NetworkReported is defined with two variables but used with three
    in the second rule:

    NetworkReported(n, s) :- RegionReported(
        r, s
    ) & RegionInNetwork(r, n)
    StudyMatchingNetworkQuery(s, n) :- (
        RegionReported("VWFA", s)
        & NetworkReported(n, s, r)
    )
    """

    pass


class TranslateToNamedRAException(NeuroLangException):
    """
    Base exception for errors translating Datalog to Named Relational Algebra.
    """

    pass


class NoValidChaseClassForStratumException(NeuroLangException):
    """
    Neurolang implements stratified datalog which splits a datalog program
    into several independent strata that can each be solved by a specific
    chase algorithm based on the properties of the rules in the stratum
    (using negation, aggregation and/or recursion).

    This exception is raised if there is no valid algorithm available to
    solve a specific stratum; e.g. no recursive compatible algorithm was
    provided to solve a recursive stratum.

    See `neurolang.datalog.chase.__init__.py` for available chase
    implementations.
    """

    pass


class CouldNotTranslateConjunctionException(TranslateToNamedRAException):
    """
    This conjunctive formula could not be translated into an equivalent
    relational algebra representation. This is probably because the
    formula is not in *modified relational algebra normal form*.

    Generaly speaking, the formula must be expressed in *conjunctive normal
    form* (CNF) or *disjunctive normal form* (DNF): as either a conjunction of
    disjunctions or disjunction of conjunctions.

    See 5.4.7 from [1]_.

    Examples
    --------
    PositiveReverseInferenceSegregationQuery(
        t, n, PROB(t, n)
    ) :- (TopicAssociation(t, s) // SelectedStudy(s)) // (
        StudyMatchingNetworkQuery(s, n) & SelectedStudy(s)
    )

    This formula is not in DNF since it is a disjunction of a disjunction
    (TopicAssociation(t, s) // SelectedStudy(s)) and a conjunction
    (StudyMatchingNetworkQuery(s, n) & SelectedStudy(s)).
    A valid query would be :

    PositiveReverseInferenceSegregationQuery(
        t, n, PROB(t, n)
    ) :- (TopicAssociation(t, s) & SelectedStudy(s)) // (
        StudyMatchingNetworkQuery(s, n) & SelectedStudy(s)
    )

    .. [1] S. Abiteboul, R. Hull, V. Vianu, Foundations of databases
        (Addison Wesley, 1995), Addison-Wesley.
    """

    def __init__(self, output):
        super().__init__(f"Could not translate conjunction: {output}")
        self.output = output


class NegativeFormulaNotSafeRangeException(TranslateToNamedRAException):
    """
    This rule is not *range restricted* and cannot be solved in
    *nonrecursive datalog with negation*. One of the variables in this rule
    appears in a negated literal without also appearing in a non-negated
    literal.

    A datalog rule composed of literals of the form R(v) or ¬R(v) is
    *range restricted* if each variable x occurring in the rule occurs in at
    least one literal of the form R(v) (non-negated literal) in the rule body.
    See 5.2 from [1]_.

    Examples
    --------
    StudyNotMatchingSegregationQuery(s, n) :- (
        ~StudyMatchingNetworkQuery(s, n)
        & Network(n)
    )

    Variable `s` is present in the negated `StudyMatchingNetworkQuery`
    literal but is not present in a non-negated literal. A valid query body
    would be :

    StudyNotMatchingSegregationQuery(s, n) :- (
        ~StudyMatchingNetworkQuery(s, n)
        & Study(s) & Network(n)
    )

    .. [1] S. Abiteboul, R. Hull, V. Vianu, Foundations of databases
        (Addison Wesley, 1995), Addison-Wesley.
    """

    def __init__(self, formula):
        super().__init__(f"Negative predicate {formula} is not safe range")
        self.formula = formula


class NegativeFormulaNotNamedRelationException(TranslateToNamedRAException):
    """
    This rule contains a negative literal R(v) which was not previously
    defined as a non-negated relation.

    Examples
    --------
    t(x, y) :- r(x, y) & q(y, z)
    s(x, y, prob(x, y)) :- ~t(x, x) & q(x, y)
    """

    def __init__(self, formula):
        super().__init__(f"Negative formula {formula} is not a named relation")
        self.formula = formula


class NonLiftableException(UnsupportedSolverError):
    pass


class NotRankedException(NonLiftableException):
    pass


class NotUnateException(NonLiftableException):
    pass

class InvalidCommandExpression(ForbiddenExpressionError):
    """
    Invalid Command statement.
    """

    pass


class ParserError(NeuroLangException):
    def __init__(self, message, line=None, column=None):
        self.message = message
        self.line = line
        self.column = column

    def __str__(self):
        return self.message


class UnexpectedTokenError(ParserError):
    pass


class UnexpectedCharactersError(ParserError):
    pass


from functools import reduce
import logging

from ..expression_walker import ChainedWalker, ReplaceExpressionWalker
from ..logic import Conjunction, Disjunction, ExistentialPredicate, Negation
from ..logic.expression_processing import (
    extract_logic_atoms,
    extract_logic_free_variables,
    extract_logic_predicates
)
from ..logic.transformations import (
    CollapseConjunctions,
    CollapseDisjunctions,
    DistributeConjunctions,
    DistributeDisjunctions,
    GuaranteeConjunction,
    GuaranteeDisjunction,
    MoveNegationsToAtomsInFONegE,
    PushExistentialsDown,
    RemoveDuplicatedConjunctsDisjuncts,
    RemoveTrivialOperations,
    RemoveUniversalPredicates,
    convert_to_pnf_with_dnf_matrix
)
from ..logic.unification import compose_substitutions, most_general_unifier
from .containment import is_contained


LOG = logging.getLogger(__name__)


GC = GuaranteeConjunction()
GD = GuaranteeDisjunction()
MNA = MoveNegationsToAtomsInFONegE()
PED = PushExistentialsDown()
RTO = RemoveTrivialOperations()


def minimize_ucq_in_cnf(query):
    """Convert UCQ to CNF form
    and minimise.

    Parameters
    ----------
    query : LogicExpression.
        query in UCQ semantics.

    Returns
    -------
    LogicExpression
        minimised query in UCQ semantics.
    """
    query = convert_to_cnf_existential_ucq(query)
    head_variables = extract_logic_free_variables(query)
    cq_d_min = Conjunction(tuple(
        minimize_component_disjunction(c, head_variables)
        for c in query.formulas
    ))

    simplify = ChainedWalker(
        MoveNegationsToAtomsInFONegE,
        PushExistentialsDown,
        RemoveDuplicatedConjunctsDisjuncts,
        RemoveTrivialOperations,
        GuaranteeConjunction,
    )

    cq_min = minimize_component_conjunction(cq_d_min, head_variables)
    cq_min = add_existentials_except(cq_min, head_variables)
    return simplify.walk(cq_min)


def minimize_ucq_in_dnf(query):
    """Convert UCQ to DNF form
    and minimise.

    Parameters
    ----------
    query : LogicExpression.
        query in UCQ semantics.

    Returns
    -------
    LogicExpression
        minimised query in UCQ semantics.
    """
    query = convert_to_dnf_existential_ucq(query)
    head_variables = extract_logic_free_variables(query)
    cq_d_min = Disjunction(tuple(
        minimize_component_conjunction(c, head_variables)
        for c in query.formulas
    ))

    simplify = ChainedWalker(
        MoveNegationsToAtomsInFONegE,
        PushExistentialsDown,
        RemoveDuplicatedConjunctsDisjuncts,
        RemoveTrivialOperations,
        GuaranteeDisjunction
    )

    cq_min = minimize_component_disjunction(cq_d_min, head_variables)
    cq_min = add_existentials_except(cq_min, head_variables)
    return simplify.walk(cq_min)


def convert_rule_to_ucq(implication):
    """Convert datalog rule to logic UCQ.
    A UCQ is defined by a logic expression in
    where the only quantifier is existential.

    Parameters
    ----------
    expression : Implication
        Datalog rule.

    Returns
    -------
    LogicExpression
       UCQ with the same ground set as the
       input datalog rule.
    """
    implication = RTO.walk(implication)
    consequent, antecedent = implication.unapply()
    head_vars = set(consequent.args)
    existential_vars = (
        extract_logic_free_variables(antecedent) -
        set(head_vars)
    )
    for a in existential_vars:
        antecedent = ExistentialPredicate(a, antecedent)
    return RTO.walk(PED.walk(antecedent))


def convert_to_cnf_existential_ucq(expression):
    """Convert logic UCQ to
    conjunctive normal from (CNF)
    with only existential quantifiers.

    Parameters
    ----------
    expression : LogicExpression
        UCQ.

    Returns
    -------
    LogicExpression
       equivalent UCQ in CNF form.
    """
    expression = RTO.walk(expression)
    expression = Conjunction((expression,))
    c = ChainedWalker(
        MoveNegationsToAtomsInFONegE,
        PushExistentialsDown,
        DistributeDisjunctions,
        CollapseConjunctions,
        CollapseDisjunctions,
    )
    return c.walk(expression)


def convert_to_dnf_existential_ucq(expression):
    """Convert logic UCQ to
    disjunctive normal from (DNF)
    with only existential quantifiers.

    Parameters
    ----------
    expression : LogicExpression
        UCQ.

    Returns
    -------
    LogicExpression
       equivalent UCQ in DNF form.
    """
    expression = RTO.walk(expression)
    expression = Disjunction((expression,))
    c = ChainedWalker(
        MoveNegationsToAtomsInFONegE,
        PushExistentialsDown,
        DistributeConjunctions,
        CollapseDisjunctions,
        CollapseConjunctions,
    )
    return c.walk(expression)


def minimize_component_disjunction(disjunction, head_vars=None):
    """Given a disjunction of queries Q1  ∨ ... ∨ Qn
    remove each query Qi such that exists Qj and
    Qi → Qj.

    Parameters
    ----------
    disjunction : Disjunction
        Disjunction of logical formulas to minimise.

    head_vars: set
        variables to be considered as constants

    Returns
    -------
    Disjunction
        Minimised disjunction.
    """

    if head_vars is None:
        head_vars = {}

    if not isinstance(disjunction, Disjunction):
        return disjunction
    positive_formulas, negative_formulas = \
        split_positive_negative_formulas(disjunction)
    keep = minimise_formulas_containment(
        positive_formulas,
        is_contained,
        head_vars
    ) + tuple(negative_formulas)

    return GD.walk(RTO.walk(Disjunction(keep)))


def minimize_component_conjunction(conjunction, head_vars=None):
    """Given a conjunction of queries Q1 ∧ ... ∧ Qn
    remove each query Qi such that exists Qj and
    Qj → Qi.

    Parameters
    ----------
    conjunction : Conjunction
        conjunction of logical formulas to minimise.

    head_vars: set
        variables to be considered as constants

    Returns
    -------
    Conjunction
        minimised conjunction.
    """

    LOG.debug("About to minimize conjunction %s", conjunction)

    if head_vars is None:
        head_vars = {}

    if not isinstance(conjunction, Conjunction):
        return conjunction
    positive_formulas, negative_formulas = \
        split_positive_negative_formulas(conjunction)
    keep = minimise_formulas_containment(
        positive_formulas,
        lambda x, y: is_contained(y, x),
        head_vars
    ) + tuple(negative_formulas)

    return GC.walk(RTO.walk(Conjunction(keep)))


def split_positive_negative_formulas(nary_logic_operation):
    """Split formulas of the n_ary_logic operation in those
    containing a negated predicate and those not.

    Parameters
    ----------
    nary_logic_operation : NAryLogicOperation
        Operation whose formulas are going to be split

    Returns
    -------
    positive, negative
        two Iterable[Union[LogicOperation, FunctionApplication]] objects
        containing the positive and negative formulas
    """

    formulas = nary_logic_operation.formulas
    positive_formulas = []
    negative_formulas = []
    for formula in formulas:
        if any(
            isinstance(sub_formula, Negation)
            for sub_formula in extract_logic_predicates(formula)
        ):
            negative_formulas.append(formula)
        else:
            positive_formulas.append(formula)
    return positive_formulas, negative_formulas


def minimise_formulas_containment(components, containment_op, head_vars):
    components_fv = [
        extract_logic_free_variables(c)
        for c in components
    ]
    keep = tuple()
    containments = {}
    for i, c in enumerate(components):
        for j, c_ in enumerate(components):
            if i == j:
                continue
            c_fv = components_fv[i] & components_fv[j]
            q = add_existentials_except(c, c_fv | head_vars)
            q_ = add_existentials_except(c_, c_fv | head_vars)
            is_contained = containments.setdefault(
                (i, j), containment_op(q_, q)
            )
            LOG.debug("Checking containment %s <= %s: %s", q, q_, is_contained)
            if (
                is_contained and
                not (
                    j < i and
                    containments.get((j, i), False)
                )
            ):
                break
        else:
            keep += (c,)
    return keep


def add_existentials_except(query, variables):
    """Existentially-quantify each free variable in query
    except for those in variables

    Parameters
    ----------
    query : LogicExpression
        logic expression to add the existential quantifiers to.
    variables : Iterable of Symbol
        variables to exclude from existential quantification.

    Returns
    -------
    LogicExpression
        logic expression with existentially-quantified variables
        added.
    """
    fv = extract_logic_free_variables(query) - variables
    for v in fv:
        query = ExistentialPredicate(v, query)
    return query


def unify_existential_variables(query):
    """Reduce the number of existentially-quantified variables.
    Specifically if query is an UCQ and can be rewritten in DNF such that
    Q = ∃x.Q1 ∨ ∃y.Q2 and x in Q1 unifies with y in
    Q2, then Q is transformed to Q = ∃f.(Q1[f/x] ∨ Q2[f/y])
    where f is possibly a fresh variable.

    Parameters
    ----------
    query : LogicExpression
        UCQ expression to unify existential variables if possible.

    Returns
    -------
    LogicExpression
        logic expression with existential variables unifies
    """
    original_query = query
    query = convert_to_pnf_with_dnf_matrix(query)
    query = RTO.walk(RemoveUniversalPredicates().walk(query))
    variables_to_project = extract_logic_free_variables(query)
    while isinstance(query, ExistentialPredicate):
        query = query.body
    if not isinstance(query, Disjunction):
        return original_query

    unifiers = []
    for i, clause in enumerate(query.formulas):
        atoms = extract_logic_atoms(clause)
        for clause_ in query.formulas[i + 1:]:
            atoms_ = extract_logic_atoms(clause_)
            unifiers += [
                most_general_unifier(a, a_)
                for a in atoms
                for a_ in atoms_
            ]
    unifiers = reduce(
        compose_substitutions,
        (u[0] for u in unifiers if u is not None),
        {}
    )
    unifiers = {
        k: v for k, v in unifiers.items()
        if variables_to_project.isdisjoint((k, v))
    }
    for i in range(len(unifiers)):
        query = ReplaceExpressionWalker(unifiers).walk(query)
    query = add_existentials_except(query, variables_to_project)
    return query



import pytest
from pathlib import Path
from neurolang import NeurolangDL
from neurolang.frontend.probabilistic_frontend import NeurolangPDL
from ..engines import (
    NeurolangEngineSet,
    NeurosynthEngineConf,
)


@pytest.fixture
def engine():
    return NeurolangDL()


def test_neurolang_engine_set(engine):
    """
    NeurolangEngineSet should manage engines with a locking semaphore
    to control that threads cannot acquire more engines than are
    available
    """
    es = NeurolangEngineSet(engine)
    assert es.counter == 1
    assert es.sema.acquire()

    # try to get engine, wait at most 1s
    with es.engine(timeout=1) as engine_:
        # should not return engine since we haven't released semaphore
        assert engine_ is None

    es.sema.release()
    with es.engine() as engine_:
        assert engine_ is engine

    # test that resources are released even when an exception is raised
    try:
        with es.engine() as engine_:
            raise RuntimeError("Oops an error occured")
    except RuntimeError:
        pass
    assert len(es.engines) == 1
    assert es.sema.acquire()


def test_neurosynth_engine_configuration():
    """
    NeurosynthEngineConf should have a unique key == "neurosynth"
    and `create` method should return an instance of `NeurolangPDL` with
    `PeakReported`, `Study`, `TermInStudyTFIDF`, `SelectedStudy` & `Voxel`
    symbols.
    """
    data_dir = Path("neurolang_data")
    conf = NeurosynthEngineConf(data_dir)
    assert conf.key == "neurosynth"

    engine = conf.create()
    assert isinstance(engine, NeurolangPDL)
    assert "PeakReported" in engine.symbols
    assert "Study" in engine.symbols
    assert "TermInStudyTFIDF" in engine.symbols
    assert "SelectedStudy" in engine.symbols
    assert "Voxel" in engine.symbols



"""
Compiler for the intermediate representation of a Datalog program.
The :class:`DatalogProgram` class processes the
intermediate representation of a program and extracts
the extensional, intensional, and builtin
sets.
"""

import operator as op
from itertools import tee
from typing import AbstractSet, Any, Callable, Tuple
from warnings import warn

from ..exceptions import (
    NotConjunctiveExpression,
    NotConjunctiveExpressionNestedPredicates,
    ProtectedKeywordError
)
from ..expression_walker import ExpressionWalker, PatternWalker, add_match
from ..expressions import (
    Constant,
    Expression,
    FunctionApplication,
    NeuroLangException,
    Symbol,
    TypedSymbolTableMixin,
    is_leq_informative
)
from ..type_system import Unknown, get_args, infer_type
from .expression_processing import (
    extract_logic_free_variables,
    is_conjunctive_expression,
    is_conjunctive_expression_with_nested_predicates
)
from .expressions import (
    NULL,
    UNDEFINED,
    Fact,
    Implication,
    NullConstant,
    Undefined,
    Union
)
from .wrapped_collections import WrappedRelationalAlgebraSet

__all__ = [
    "Implication", "Fact", "Undefined", "NullConstant",
    "UNDEFINED", "NULL", "WrappedRelationalAlgebraSet",
    "DatalogProgram", "UnionOfConjunctiveQueries"
]


class UnionOfConjunctiveQueries:
    pass


class DatalogProgramMixin(TypedSymbolTableMixin, PatternWalker):
    '''
    Implementation of Datalog grammar in terms of
    Intermediate Representations. No query resolution implemented.
    In the symbol table the value for a symbol `S` is implemented as:

    * If `S` is part of the extensional database, then the value of the symbol
    is a set of tuples `a` representing `S(*a)` as facts. The type of the
    symbol must be more informative than `AbstractSet[Tuple]`

    * If `S` is part of the intensional database then its value is an
    `Union` of `Implications`. For instance
    `Q(x) :- R(x, x)` and `Q(x) :- T(x)` is represented as a symbol `Q`
     with value
     `Union((Implication(Q(x), R(x, x)), Implication(Q(x), T(x))))`.
     The type of the symbol `Q` in the symbol_table will be
     `UnionOfConjunctiveQueries`.
    '''

    def __init_subclass__(cls, **kwargs):
        for c in cls.mro()[1:-1]:
            if not hasattr(c, 'protected_keywords'):
                continue
            cls.protected_keywords = (
                cls.protected_keywords |
                c.protected_keywords
            )
        super().__init_subclass__(**kwargs)

    protected_keywords = set()

    constant_equals = Constant[Callable[[Any, Any], bool]](op.eq)

    @add_match(Symbol)
    def symbol(self, expression):
        new_expression = self.symbol_table.get(expression, expression)
        if new_expression is expression:
            return expression
        elif isinstance(new_expression, (Union, Constant[AbstractSet])):
            return expression
        else:
            return new_expression

    @add_match(Union)
    def union_of_cq(self, expression):
        """This pattern is here to avoid
        processing the full program when one rule
        is rewritten.
        """
        for formula in expression.formulas:
            self.walk(formula)

        return expression

    @add_match(Fact(FunctionApplication(Symbol, ...)))
    def fact(self, expression):
        fact = expression.fact
        if fact.functor.name in self.protected_keywords:
            raise ProtectedKeywordError(
                f'symbol {self.constant_set_name} is protected'
            )

        if any(
            not isinstance(a, Constant)
            for a in fact.args
        ):
            raise NeuroLangException(
                'Facts can only have constants as arguments'
            )

        self._initialize_fact_set_if_needed(fact)
        fact_set = self.symbol_table[fact.functor]

        if isinstance(fact_set, Union):
            raise NeuroLangException(
                f'{fact.functor} has been previously '
                'define as intensional predicate.'
            )

        fact_set.value.add(fact.args)

        return expression

    def _initialize_fact_set_if_needed(self, fact):
        if fact.functor not in self.symbol_table:
            if fact.functor.type is Unknown:
                c = Constant(fact.args)
                set_type = c.type
            elif isinstance(fact.functor.type, Callable):
                set_type = Tuple[fact.functor.type.__args__[:-1]]
            else:
                raise NeuroLangException('Fact functor type incorrect')

            self.symbol_table[fact.functor] = \
                Constant[AbstractSet[set_type]](
                    WrappedRelationalAlgebraSet(),
                    verify_type=False
                )

    @add_match(Implication(
        FunctionApplication[bool](Symbol, ...),
        Expression
    ))
    def statement_intensional(self, expression):
        original_expression = expression
        consequent = self.walk(expression.consequent)
        antecedent = self.walk(expression.antecedent)
        expression = Implication(consequent, antecedent)

        self._validate_implication_syntax(consequent, antecedent)

        symbol = consequent.functor.cast(UnionOfConjunctiveQueries)

        if symbol in self.symbol_table:
            disj = self._new_intensional_internal_representation(consequent)
        else:
            disj = tuple()

        if expression not in disj:
            disj += (expression,)

        self.symbol_table[symbol] = Union(disj)

        return original_expression

    def _new_intensional_internal_representation(self, consequent):
        symbol = consequent.functor.cast(UnionOfConjunctiveQueries)
        value = self.symbol_table[symbol]
        if (
            isinstance(value, Constant) and
            is_leq_informative(value.type, AbstractSet)
        ):
            raise NeuroLangException(
                f'{consequent.functor.name} has been previously '
                'defined as Fact or extensional database.'
            )
        disj = self.symbol_table[symbol].formulas

        if (
            not isinstance(disj[0].consequent, FunctionApplication) or
            len(extract_logic_free_variables(disj[0].consequent.args)) !=
            len(consequent.args)
        ):
            raise NeuroLangException(
                f"{disj[0].consequent} is already in the IDB "
                f"with different signature."
            )

        return disj

    def _validate_implication_syntax(self, consequent, antecedent):
        if consequent.functor.name in self.protected_keywords:
            raise ProtectedKeywordError(
                f'symbol {consequent.functor} is protected'
            )

        if any(
            not isinstance(arg, (Constant, Symbol))
            for arg in consequent.args
        ):
            raise NeuroLangException(
                f'The consequent {consequent} can only be '
                'constants or symbols'
            )

        consequent_symbols = consequent._symbols - consequent.functor._symbols

        if not consequent_symbols.issubset(antecedent._symbols):
            raise NeuroLangException(
                "All variables on the consequent need to be on the antecedent"
            )

        if not is_conjunctive_expression(consequent):
            raise NotConjunctiveExpression(
                f'Expression {consequent} is not conjunctive'
            )

        if not is_conjunctive_expression_with_nested_predicates(antecedent):
            raise NotConjunctiveExpressionNestedPredicates(
                f'Expression {antecedent} is not conjunctive'
            )

    @staticmethod
    def new_set(iterable=None, row_type=None, verify_row_type=False):
        return WrappedRelationalAlgebraSet(
            iterable=iterable, row_type=row_type,
            verify_row_type=verify_row_type
        )

    def intensional_database(self):
        return {
            k: v for k, v
            in self.symbol_table.items()
            if (
                k not in self.protected_keywords
                and k.type is UnionOfConjunctiveQueries
            )
        }

    def predicate_terms(self, predicate):
        try:
            pred_repr = self.symbol_table[predicate]
            if isinstance(pred_repr, Union):
                head_args = self._predicate_terms_intensional(pred_repr)
                return head_args
            elif is_leq_informative(pred_repr.type, AbstractSet):
                row_type = get_args(pred_repr.type)[0]
                row_len = len(get_args(row_type))
                return tuple(
                    Symbol(str(i))
                    for i in range(row_len)
                )
            else:
                raise NeuroLangException(f'Predicate {predicate} not found')
        except KeyError:
            raise NeuroLangException(f'Predicate {predicate} not found')

    def _predicate_terms_intensional(self, pred_repr):
        head_args = None
        for formula in pred_repr.formulas:
            if head_args is None:
                head_args = formula.consequent.args
            elif head_args != formula.consequent.args:
                warn(
                    'Several argument names found in the rules '
                    f'defining {formula.consequent.functor.name}, keeping one'
                )
                break
        return head_args

    def extensional_database(self):
        ret = self.symbol_table.symbols_by_type(AbstractSet)
        for keyword in self.protected_keywords:
            if keyword in ret:
                del ret[keyword]
        return ret

    def builtins(self):
        return self.symbol_table.symbols_by_type(Callable)

    def add_extensional_predicate_from_tuples(
        self, symbol, iterable, type_=Unknown
    ):
        if type_ is Unknown:
            new_set = self.new_set(iterable)
            type_ = new_set.row_type
        else:
            new_set = self.new_set(iterable=iterable, row_type=type_)

        constant = Constant[AbstractSet[type_]](
            new_set,
            auto_infer_type=False,
            verify_type=False
        )
        symbol = symbol.cast(constant.type)
        self.symbol_table[symbol] = constant

    @staticmethod
    def infer_iterable_type(iterable):
        """Infer the type of iterable elements
        without modifying the iterable.

        Parameters
        ----------
        iterable : Iterable
            iterable from which to infer
            element's type

        Returns
        -------
        type, iterable
            the inferred type and the iterable.
        """
        type_ = Unknown
        if hasattr(iterable, 'fetch_one'):
            if iterable.is_empty():
                first = None
            else:
                first = iterable.fetch_one()
                if first == tuple():
                    first = None
        else:
            iterable_, iterable = tee(iterable)
            try:
                first = next(iterable_)
            except StopIteration:
                first = None
        if isinstance(first, Expression):
            type_ = first.type
        elif first is not None:
            type_ = infer_type(first)
        return type_, iterable


class DatalogProgram(DatalogProgramMixin, ExpressionWalker):
    pass



import operator
from typing import AbstractSet

import numpy as np
import pytest

from ...config import config
from ...datalog import Fact
from ...exceptions import NonLiftableException, UnsupportedSolverError
from ...expressions import Constant, Symbol
from ...logic import Conjunction, Implication, Negation, Union
from ...relational_algebra import (
    NamedRelationalAlgebraFrozenSet,
    str2columnstr_constant
)
from ...relational_algebra_provenance import ProvenanceAlgebraSet
from .. import (
    dalvi_suciu_lift,
    small_dichotomy_theorem_based_solver,
    weighted_model_counting
)
from ..cplogic import testing
from ..cplogic.program import CPLogicProgram
from ..exceptions import NotHierarchicalQueryException

try:
    from contextlib import nullcontext
except ImportError:
    from contextlib import suppress as nullcontext

EQ = Constant(operator.eq)
GT = Constant(operator.gt)
NE = Constant(operator.ne)

ans = Symbol("ans")
P = Symbol("P")
Q = Symbol("Q")
R = Symbol("R")
Z = Symbol("Z")
H = Symbol("H")
A = Symbol("A")
B = Symbol("B")
C = Symbol("C")
w = Symbol("w")
x = Symbol("x")
y = Symbol("y")
z = Symbol("z")
p = Symbol("p")

a = Constant("a")
b = Constant("b")
c = Constant("c")


@pytest.fixture(
    params=((
        weighted_model_counting,
        small_dichotomy_theorem_based_solver,
        dalvi_suciu_lift,
    )),
    ids=["SDD-WMC", "small-dichotomy", "dalvi-suciu"],
)
def solver(request):
    return request.param


def test_deterministic(solver):
    """
    We define the program

        P(x) <- Q(x)

    And we expect the provenance set resulting from the
    marginalisation of P(x) to be

        _p_ | x
        ====|===
        1.0 | a
        1.0 | b

    """
    code = Union(
        (
            Fact(Q(a)),
            Fact(Q(b)),
            Implication(P(x), Q(x)),
        )
    )
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    query = Implication(ans(x), P(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set([(1.0, "a"), (1.0, "b")], ("_p_", "x"))
    assert testing.eq_prov_relations(result, expected)


def test_deterministic_conjunction_varying_arity(solver):
    code = Union(
        (
            Fact(Q(a, b)),
            Fact(P(a)),
            Fact(P(b)),
            Implication(Z(x, y), Conjunction((Q(x, y), P(x), P(y)))),
        )
    )
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    query = Implication(ans(x, y), Z(x, y))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set([(1.0, "a", "b")], ("_p_", "x", "y"))
    assert testing.eq_prov_relations(result, expected)


def test_deterministic_conjunction_varying_arity_empty(solver):
    code = Union(
        (
            Fact(P(a)),
            Fact(P(b)),
            Implication(
                Z(x, y),
                Conjunction((Q(x, y), P(x), P(y)))
            ),
        )
    )
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    query = Implication(ans(x, y), Z(x, y))
    result = solver.solve_succ_query(query, cpl_program)
    assert result.relation.value.is_empty()


def test_simple_bernoulli(solver):
    """
    We define the program

        P(a) : 0.7 <- T
        P(b) : 0.8 <- T

    And expect the provenance set resulting from the
    marginalisation of P(x) to be

        _p_ | x
        ====|===
        0.7 | a
        0.8 | b

    """
    code = Union(())
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    cpl_program.add_probabilistic_facts_from_tuples(
        P, {(0.7, "a"), (0.8, "b")}
    )
    query = Implication(ans(x), P(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set([(0.7, "a"), (0.8, "b")], ("_p_", "x"))
    assert testing.eq_prov_relations(result, expected)


def test_bernoulli_conjunction(solver):
    code = Union((Implication(Z(x), Conjunction((P(x), Q(x), R(x)))),))
    probfacts_sets = {
        P: {(1.0, "a"), (0.5, "b")},
        Q: {(0.9, "b"), (0.1, "c")},
        R: {(0.9, "b"), (0.1, "c")},
    }
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    for pred_symb, pfact_set in probfacts_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    query = Implication(ans(x), Z(x))
    result = solver.solve_succ_query(query, cpl_program)

    expected = testing.make_prov_set([(0.9 * 0.9 * 0.5, "b")], ("_p_", "x"))
    assert testing.eq_prov_relations(result, expected)


def test_probfact_existential():
    cpl_program = CPLogicProgram()
    cpl_program.add_probabilistic_facts_from_tuples(
        P, {(0.7, 1, 2), (0.9, 1, 3), (0.88, 2, 4)}
    )
    query = Implication(ans(x), P(x, y))
    result = dalvi_suciu_lift.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            ((1 - (1 - 0.7) * (1 - 0.9)), 1),
            (0.88, 2),
        ],
        ("_p_", "x"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_multi_level_conjunction(solver):
    """
    We consider the program

           P(a) : 0.2 <- T
           Q(a) : 0.9 <- T
        R(a, a) : 0.1 <- T
        R(a, b) : 0.5 <- T
                 Z(x) <- P(x), Q(x)
              H(x, y) <- Z(x), R(x, y)

    And expect the prov set resulting from the
    marginalisation of H(x, y) to be

        _p_   | x  | y
        ======|====|===
        0.018 | a  | a
        0.09  | a  | b

    """
    probfacts_sets = {
        P: {(0.2, "a")},
        Q: {(0.9, "a")},
        R: {(0.1, "a", "a"), (0.5, "a", "b")},
    }
    code = Union(
        (
            Implication(Z(x), Conjunction((P(x), Q(x)))),
            Implication(H(x, y), Conjunction((Z(x), R(x, y)))),
        )
    )
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    for pred_symb, pfact_set in probfacts_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    query = Implication(ans(x, y), H(x, y))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.2 * 0.9 * 0.1, "a", "a"),
            (0.2 * 0.9 * 0.5, "a", "b"),
        ],
        ("_p_", "x", "y"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_intertwined_conjunctions_and_probfacts(solver):
    """
    We consider the program

        P(a) : 0.8  <-  T
        C(a) : 0.5  <-  T
        C(b) : 0.9  <-  T
              A(x)  <-  B(x), C(x)
              B(x)  <-  P(x)
              Z(x)  <-  A(x), B(x), C(x)

    And expect SUCC[ Z(x) ] to yield the provenance relation

        _p_ | x
        ====|===
        0.4 | a

    """
    cpl_code = Union(
        (
            Implication(A(x), Conjunction((B(x), C(x)))),
            Implication(B(x), P(x)),
            Implication(Z(x), Conjunction((A(x), B(x), C(x)))),
        )
    )
    cpl = CPLogicProgram()
    cpl.walk(cpl_code)
    cpl.add_probabilistic_facts_from_tuples(P, {(0.8, "a")})
    cpl.add_probabilistic_facts_from_tuples(C, {(0.5, "a"), (0.9, "b")})
    query = Implication(ans(y), Z(y))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set([(0.8 * 0.5, "a")], ("_p_", "y"))
    assert testing.eq_prov_relations(result, expected)


def test_simple_probchoice(solver):
    pchoice_as_sets = {P: {(0.2, "a"), (0.8, "b")}}
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    query = Implication(ans(x), P(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.2, "a"),
            (0.8, "b"),
        ],
        ("_p_", "x"),
    )
    assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_mutual_exclusivity(solver):
    pchoice_as_sets = {P: {(0.2, "a"), (0.8, "b")}}
    pfact_sets = {Q: {(0.5, "a", "b")}}
    code = Union((Implication(Z(x, y), Conjunction((P(x), P(y), Q(x, y)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(x, y), Z(x, y))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set([], ("_p_", "x", "y"))
    assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_fact_negation_choice(solver):
    pchoice_as_sets = {P: {(0.2, "a"), (0.8, "b")}}
    pfact_sets = {Q: {(0.5, "a", "b")}}
    code = Union((
        Implication(
            Z(x, y),
            Conjunction((Negation(P(y)), Q(x, y)))
        ),
    ))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(x, y), Z(x, y))

    if solver is small_dichotomy_theorem_based_solver:
        context = pytest.raises(NotHierarchicalQueryException)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl_program)
        expected = testing.make_prov_set([(0.1, "a", "b")], ("_p_", "x", "y"))
        assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
    pytest.param(dalvi_suciu_lift, marks=pytest.mark.xfail(
        reason="Connected component issue addressed by PR #660"
    )),
])
def test_negation_fact_choice(solver):
    pchoice_as_sets = {P: {(0.2, "a"), (0.8, "b")}}
    pfact_sets = {Q: {(0.4, "a", "a")}}
    code = Union((
        Implication(
            Z(x),
            Conjunction((P(x), Negation(Q(x, x))))
        ),
    ))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(x), Z(x))

    if solver is small_dichotomy_theorem_based_solver:
        context = pytest.raises(NotHierarchicalQueryException)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl_program)
        expected = testing.make_prov_set([(.12, "a"), (.8, "b")], ("_p_", "x"))
        assert testing.eq_prov_relations(result, expected)


def test_multiple_probchoices_mutual_exclusivity(solver):
    pchoice_as_sets = {
        P: {(0.2, "a"), (0.8, "b")},
        Q: {(0.5, "a", "b"), (0.4, "b", "c"), (0.1, "b", "b")},
    }
    rule = Implication(Z(x, y), Conjunction((P(x), Q(y, y))))
    code = Union((rule,))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    cpl_program.walk(code)
    query = Implication(ans(x, y), Z(x, y))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [(0.2 * 0.1, "a", "b"), (0.8 * 0.1, "b", "b")], ("_p_", "x", "y")
    )
    assert testing.eq_prov_relations(result, expected)


def test_large_probabilistic_choice(solver):
    if (
        config["RAS"].get("backend", "pandas") == "dask"
        or solver is weighted_model_counting
    ):
        n = int(1000)
    else:
        n = int(10000)
    with testing.temp_seed(42):
        probs = np.random.rand(n)
    probs = probs / probs.sum()
    pchoice_as_sets = {P: {(float(prob), i) for i, prob in enumerate(probs)}}
    pfact_sets = {Q: {(0.5, 0, 0), (0.5, 0, 1)}}
    code = Union((Implication(Z(x, y), Conjunction((P(x), Q(x, y)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(x, y), Z(x, y))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.5 * probs[0], 0, 0),
            (0.5 * probs[0], 0, 1),
        ],
        ("_p_", "x", "y"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_simple_probchoice_existential(solver):
    """
    We define the following program

        P(a, a) : 0.2 v P(a, b) : 0.7 v P(c, c) : 0.1 <- T
        Q(x) <- ∃y, P(x, y)

    We expect the following to hold

        - Pr[Q(a)] = 0.9
        - Pr[Q(c)] = 0.1

    """
    pchoice_as_sets = {P: {(0.2, "a", "a"), (0.7, "a", "b"), (0.1, "c", "c")}}
    code = Union((Implication(Q(x), P(x, y)),))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    cpl_program.walk(code)
    query = Implication(ans(x), Q(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set([(0.9, "a"), (0.1, "c")], ("_p_", "x"))
    assert testing.eq_prov_relations(result, expected)


def test_existential_in_conjunction(solver):
    pchoice_as_sets = {
        P: {(0.2, "a", "b", "c"), (0.4, "b", "b", "c"), (0.4, "b", "a", "c")},
        Z: {(0.5, "b"), (0.5, "d")},
    }
    code = Union((Implication(Q(x), Conjunction((Z(y), P(x, y, z)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    cpl_program.walk(code)
    query = Implication(ans(x), Q(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set([(0.1, "a"), (0.2, "b")], ("_p_", "x"))
    assert testing.eq_prov_relations(result, expected)


def test_existential_alternative_variables(solver):
    pchoice_as_sets = {
        P: {(0.8, "a", "b"), (0.1, "c", "d"), (0.1, "d", "e")},
    }
    pfact_sets = {
        Z: {(0.2, "a"), (0.7, "e")},
    }
    code = Union(
        (
            Fact(R(Constant[str]("a"))),
            Fact(R(Constant[str]("b"))),
            Implication(H(x), Conjunction((Z(y), P(y, x)))),
        )
    )
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(z), H(z))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set([(0.2 * 0.8, "b")], ("prob", "z"))
    assert testing.eq_prov_relations(result, expected)


def test_multilevel_existential(solver):
    pchoice_as_sets = {
        P: {(0.5, "a", "b"), (0.5, "b", "c")},
        R: {(0.1, "a"), (0.4, "b"), (0.5, "c")},
        Q: {(0.9, "a"), (0.1, "c")},
        Z: {(0.1, "b"), (0.9, "c")},
    }
    code = Union(
        (
            Implication(H(x, y), Conjunction((R(x), Z(y)))),
            Implication(A(x), Conjunction((H(x, y), P(y, x)))),
            Implication(B(x), Conjunction((A(x), Q(y)))),
            Implication(C(x), H(x, y)),
        )
    )
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    cpl_program.walk(code)
    query = Implication(ans(x, y), H(x, y))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.1 * 0.1, "a", "b"),
            (0.1 * 0.9, "a", "c"),
            (0.4 * 0.1, "b", "b"),
            (0.4 * 0.9, "b", "c"),
            (0.5 * 0.1, "c", "b"),
            (0.5 * 0.9, "c", "c"),
        ],
        ("_p_", "x", "y"),
    )
    assert testing.eq_prov_relations(result, expected)

    query = Implication(ans(z), C(z))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.1, "a"),
            (0.4, "b"),
            (0.5, "c"),
        ],
        ("_p_", "z"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_multilevel_existential_2(solver):
    pchoice_as_sets = {
        P: {(0.5, "a", "b"), (0.5, "b", "c")},
        R: {(0.1, "a"), (0.4, "b"), (0.5, "c")},
        Q: {(0.9, "a"), (0.1, "c")},
        Z: {(0.1, "b"), (0.9, "c")},
    }
    code = Union(
        (
            Implication(H(x, y), Conjunction((R(x), Z(y)))),
            Implication(A(x), Conjunction((H(x, y), P(y, x)))),
            Implication(B(x), Conjunction((A(x), Q(y)))),
            Implication(C(x), H(x, y)),
        )
    )
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    cpl_program.walk(code)

    query = Implication(ans(z), B(z))

    if solver is small_dichotomy_theorem_based_solver:
        context = pytest.raises(NotHierarchicalQueryException)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl_program)
        expected = testing.make_prov_set(
            [(0.5 * 0.1 * 0.5, "c")],
            ("_p_", "z"),
        )
        assert testing.eq_prov_relations(result, expected)


def test_repeated_antecedent_predicate_symbol(solver):
    """
    We consider the simple program

        P(a) : 0.4  <-  T
        P(b) : 0.7  <-  T
           Q(x, y)  <-  P(x), P(y)

    Possible outcomes are

        { }                     with prob   (1 - 0.4) * (1 - 0.7)
        { P(a), Q(a, a) }       with prob   0.4 * (1 - 0.7)
        { P(b), Q(b, b) }       with prob   0.7 * (1 - 0.4)
        { P(a), P(b),
          Q(a, a), Q(a, b),     with prob   0.4 * 0.7
          Q(b, b), Q(b, a) }

    We expected the following provenance set to result from the
    succ query prob[Q(x, y)]?

        _p_                         | x | y
        ----------------------------|---|---
        0.4 * (1 - 0.7) + 0.4 * 0.7 | a | a
        0.7 * (1 - 0.4) + 0.4 * 0.7 | b | b
        0.4 * 0.7                   | a | b
        0.4 * 0.7                   | b | a

    """
    pfact_sets = {
        P: {(0.4, "a"), (0.7, "b")},
    }
    code = Union((Implication(Q(x, y), Conjunction((P(x), P(y)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(x, y), Q(x, y))

    if solver is small_dichotomy_theorem_based_solver:
        context = pytest.raises(NotHierarchicalQueryException)
    elif solver is dalvi_suciu_lift:
        context = pytest.raises(NonLiftableException)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl_program)
        expected = testing.make_prov_set(
            [
                (0.4 * (1 - 0.7) + 0.4 * 0.7, "a", "a"),
                (0.7 * (1 - 0.4) + 0.4 * 0.7, "b", "b"),
                (0.4 * 0.7, "a", "b"),
                (0.4 * 0.7, "b", "a"),
            ],
            ("_p_", "x", "y"),
        )
        assert testing.eq_prov_relations(result, expected)


def test_tiny_cbma_example(solver):
    TermInStudy = Symbol("TermInStudy")
    ActivationReported = Symbol("ActivationReported")
    SelectedStudy = Symbol("SelectedStudy")
    TermAssociation = Symbol("TermAssociation")
    Activation = Symbol("Activation")
    s = Symbol("s")
    t = Symbol("t")
    v = Symbol("v")
    code = Union(
        (
            Implication(
                TermAssociation(t),
                Conjunction([TermInStudy(t, s), SelectedStudy(s)]),
            ),
            Implication(
                Activation(v),
                Conjunction([ActivationReported(v, s), SelectedStudy(s)]),
            ),
        )
    )
    pfact_sets = {
        TermInStudy: {
            (0.001, "memory", "1"),
            (0.002, "memory", "2"),
            (0.015, "visual", "2"),
            (0.004, "memory", "3"),
            (0.005, "visual", "4"),
            (0.0001, "memory", "4"),
            (0.01, "visual", "5"),
        },
        ActivationReported: {
            (1.0, "v1", "1"),
            (1.0, "v2", "1"),
            (1.0, "v3", "2"),
            (1.0, "v1", "3"),
            (1.0, "v1", "4"),
        },
    }
    pchoice_as_sets = {
        SelectedStudy: {
            (0.2, "1"),
            (0.2, "2"),
            (0.2, "3"),
            (0.2, "4"),
            (0.2, "5"),
        }
    }
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(t), TermAssociation(t))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (
                sum(t[0] for t in pfact_sets[TermInStudy] if t[1] == term) / 5,
                term,
            )
            for term in ("memory", "visual")
        ],
        ("_p_", "t"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_conjunct_pfact_equantified_pchoice(solver):
    pfact_sets = {P: {(0.8, "a", "s1"), (0.5, "a", "s2"), (0.1, "b", "s2")}}
    pchoice_as_sets = {Z: {(0.6, "s1"), (0.4, "s2")}}
    code = Union((Implication(Q(x), Conjunction((P(x, y), Z(y)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(x), Q(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.6 * 0.8 + 0.4 * 0.5, "a"),
            (0.1 * 0.4, "b"),
        ],
        ("_p_", "x"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_shatterable_query(solver):
    pfact_sets = {P: {(0.8, "a", "1"), (0.5, "a", "2"), (0.1, "b", "2")}}
    code = Union((Implication(Q(x), Conjunction((P(a, x), P(b, x)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(x), Q(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.5 * 0.1, "2"),
        ],
        ("_p_", "x"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_shatterable_query_2(solver):
    pfact_sets = {
        P: {(0.8, "a", "c", "1"), (0.5, "a", "c", "2"), (0.1, "b", "b", "2")}
    }
    code = Union((Implication(Q(x), Conjunction((P(a, c, x), P(b, b, x)))),))
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    cpl_program.walk(code)
    query = Implication(ans(x), Q(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.5 * 0.1, "2"),
        ],
        ("_p_", "x"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_program_with_variable_equality(solver):
    pfact_sets = {
        Q: {(0.2, "a"), (0.3, "b"), (0.4, "c")},
    }
    pchoice_as_sets = {
        P: {(0.5, "a", "b"), (0.5, "b", "c")},
    }
    code = Union(
        (Implication(Z(x, y), Conjunction((P(y, x), Q(y), EQ(y, a)))),)
    )
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfact_set
        )
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    cpl_program.walk(code)
    query = Implication(ans(x, y), Z(x, y))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.5 * 0.2, "b", "a"),
        ],
        ("_p_", "x", "y"),
    )
    assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_program_with_segregation_constant_case(solver):
    pfact_sets = {
        P: {(0.5, "a", "b"), (0.5, "b", "c"), (0.2, "b", "d")},
    }
    code = Union((
        Implication(Z(x, y), Conjunction((P(x, y), Negation(A(x, y))))),
        Implication(A(x, y), Conjunction((P(w, y), NE(w, x)))),
    ))
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfact_set
        )
    cpl_program.walk(code)
    query = Implication(ans(), Z(a, b))

    if solver is small_dichotomy_theorem_based_solver:
        context = pytest.raises(UnsupportedSolverError)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl_program)
        expected = testing.make_prov_set(
            [
                (0.5,),
            ],
            ("_p_",),
        )
        assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_program_with_segregation_first_constant(solver):
    pfact_sets = {
        P: {(0.5, "a", "b"), (0.5, "b", "c"), (0.2, "b", "d")},
    }
    code = Union((
        Implication(Z(x, y), Conjunction((P(x, y), Negation(A(x, y))))),
        Implication(A(x, y), Conjunction((P(w, y), NE(w, x)))),
    ))
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfact_set
        )
    cpl_program.walk(code)
    query = Implication(ans(y), Z(a, y))

    if solver is small_dichotomy_theorem_based_solver:
        context = pytest.raises(UnsupportedSolverError)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl_program)
        expected = testing.make_prov_set(
            [
                (0.5, "b"),
            ],
            ("_p_", "y"),
        )
        assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_program_with_segregation_second_constant(solver):
    pfact_sets = {
        P: {(0.5, "a", "b"), (0.5, "b", "c"), (0.2, "b", "d")},
    }
    code = Union((
        Implication(Z(x, y), Conjunction((P(x, y), Negation(A(x, y))))),
        Implication(A(x, y), Conjunction((P(w, y), NE(w, x)))),
    ))
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfact_set
        )
    cpl_program.walk(code)
    query = Implication(ans(x), Z(x, b))

    if (
        solver is small_dichotomy_theorem_based_solver or
        solver is dalvi_suciu_lift
    ):
        context = pytest.raises(UnsupportedSolverError)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl_program)
        expected = testing.make_prov_set(
            [
                (0.5, "a"),
            ],
            ("_p_", "x"),
        )
        assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_program_with_segregation(solver):
    pfact_sets = {
        P: {(0.5, "a", "b"), (0.5, "b", "c"), (0.2, "b", "d")},
    }
    code = Union((
        Implication(Z(x, y), Conjunction((P(x, y), Negation(A(x, y))))),
        Implication(A(x, y), Conjunction((P(w, y), NE(w, x)))),
    ))
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfact_set
        )
    cpl_program.walk(code)
    query = Implication(ans(x, y), Z(x, y))

    if (
        solver is small_dichotomy_theorem_based_solver or
        solver is dalvi_suciu_lift
    ):
        context = pytest.raises(UnsupportedSolverError)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl_program)
        expected = testing.make_prov_set(
            [
                (0.5, "a", "b"),
            ],
            ("_p_", "x", "y"),
        )
        assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_program_with_segregation_first_existential(solver):
    pfact_sets = {
        P: {(0.5, "a", "b"), (0.5, "b", "c"), (0.2, "b", "d")},
    }
    code = Union((
        Implication(Z(x, y), Conjunction((P(x, y), Negation(A(x, y))))),
        Implication(A(x, y), Conjunction((P(w, y), NE(w, x)))),
    ))
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfact_set
        )
    cpl_program.walk(code)
    query = Implication(ans(y), Z(x, y))

    if (
        solver is small_dichotomy_theorem_based_solver or
        solver is dalvi_suciu_lift
    ):
        context = pytest.raises(UnsupportedSolverError)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl_program)
        expected = testing.make_prov_set(
            [
                (0.5, "b"),
            ],
            ("_p_", "y"),
        )
        assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_program_with_segregation_second_existential(solver):
    pfact_sets = {
        P: {(0.5, "a", "b"), (0.5, "b", "c"), (0.2, "b", "d")},
    }
    code = Union((
        Implication(Z(x, y), Conjunction((P(x, y), Negation(A(x, y))))),
        Implication(A(x, y), Conjunction((P(w, y), NE(w, x)))),
    ))
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfact_set
        )
    cpl_program.walk(code)
    query = Implication(ans(x), Z(x, y))

    if (
        solver is small_dichotomy_theorem_based_solver or
        solver is dalvi_suciu_lift
    ):
        context = pytest.raises(UnsupportedSolverError)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl_program)
        expected = testing.make_prov_set(
            [
                (0.5, "a"),
            ],
            ("_p_", "x"),
        )
        assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_program_with_negative_existential(solver):
    pfact_sets = {
        P: {(0.5, "a", "b"), (0.5, "b", "c"), (0.2, "b", "d")},
        Q: {(1., "b", "c")}
    }
    code = Union((
        Implication(Z(x, y), Conjunction((P(x, y), Negation(A(y))))),
        Implication(A(y), Q(w, y)),
    ))
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_set in pfact_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfact_set
        )
    cpl_program.walk(code)
    query = Implication(ans(x, y), Z(x, y))

    if solver is small_dichotomy_theorem_based_solver:
        context = pytest.raises(UnsupportedSolverError)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl_program)
        expected = testing.make_prov_set(
            [
                (0.5, "a", "b"),
                (0.2, "b", "d")
            ],
            ("_p_", "x", "y"),
        )
        assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    weighted_model_counting,
    pytest.param(small_dichotomy_theorem_based_solver, marks=pytest.mark.xfail(
        reason="Existential variable in probfact leads to unsupported noisy-or"
    )),
    dalvi_suciu_lift,
])
def test_repeated_variable_probabilistic_rule(solver):
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        Q, [(0.2, 7, 7, 2), (0.5, 7, 8, 4)]
    )
    cpl.walk(Implication(H(x, x), Q(x, x, y)))
    query = Implication(ans(x, y), H(x, y))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set([(0.2, 7, 7)], ("_p_", "x", "y"))
    assert testing.eq_prov_relations(result, expected)


def test_repeated_variable_with_constant_in_head(solver):
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.2, 7, 8), (0.6, 8, 9), (0.9, 8, 8)],
    )
    cpl.add_probabilistic_choice_from_tuples(
        P,
        [(0.4, 8), (0.6, 9)],
    )
    cpl.walk(
        Implication(R(Constant[int](8), x), Conjunction((Q(x, x), P(x))))
    )
    query = Implication(ans(x, y), R(x, y))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set(
        [(0.4 * 0.9, 8, 8)],
        ("_p_", "x", "y"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_empty_result_program(solver):
    rule = Implication(R(Constant(2), Constant(3)), Conjunction((Q(x),)))
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.2, 7)],
    )
    cpl.walk(rule)
    query = Implication(ans(x), R(x, x))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set([], ("_p_", "x"))
    assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_simple_negation(solver):
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.2, 'a'), (0.1, 'b')]
    )
    cpl.add_probabilistic_facts_from_tuples(
        R,
        [(0.1, 'a'), (0.3, 'c')]
    )
    rule = Implication(P(x), Conjunction((Q(x), Negation(R(x)))))
    cpl.walk(rule)
    query = Implication(ans(x), P(x))

    if solver is small_dichotomy_theorem_based_solver:
        context = pytest.raises(NotHierarchicalQueryException)
    else:
        context = nullcontext()

    with context:
        result = solver.solve_succ_query(query, cpl)
        expected = testing.make_prov_set(
            [(.18, 'a'), (.10, 'b')], ("_p_", "x")
        )
        assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_program_with_probchoice_selfjoin(solver):
    cpl = CPLogicProgram()
    cpl.add_probabilistic_choice_from_tuples(
        P,
        [(0.2, "a"), (0.8, "b")],
    )
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.6, "a"), (0.8, "b")],
    )
    rule = Implication(R(x, y), Conjunction((Q(x), P(x), P(y))))
    cpl.walk(rule)
    query = Implication(ans(x, y), R(x, y))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set(
        [(0.2 * 0.6, "a", "a"), (0.8 * 0.8, "b", "b")], ("_p_", "x", "y")
    )
    assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_probchoice_selfjoin_multiple_variables(solver):
    cpl = CPLogicProgram()
    cpl.add_probabilistic_choice_from_tuples(
        P,
        [(0.2, "a", "b"), (0.8, "b", "c")],
    )
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.6, "a"), (0.8, "b")],
    )
    rule = Implication(R(x, y, z, w), Conjunction((Q(x), P(x, y), P(z, w))))
    cpl.walk(rule)
    query = Implication(ans(x, y, z, w), R(x, y, z, w))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set(
        [(0.2 * 0.6, "a", "b", "a", "b"), (0.8 * 0.8, "b", "c", "b", "c")],
        ("_p_", "x", "y", "z", "w"),
    )
    assert testing.eq_prov_relations(result, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_probchoice_selfjoin_multiple_variables_shared_var(solver):
    cpl = CPLogicProgram()
    cpl.add_probabilistic_choice_from_tuples(
        P,
        [(0.2, "a", "b"), (0.8, "b", "b")],
    )
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.6, "a", "b"), (0.8, "b", "a"), (0.2, "b", "b")],
    )
    rule = Implication(R(x, y, z), Conjunction((Q(x, y), P(x, y), P(z, x))))
    cpl.walk(rule)
    query = Implication(ans(x, y, z), R(x, y, z))
    result = solver.solve_succ_query(query, cpl)
    expected = testing.make_prov_set(
        [(0.2 * 0.8, "b", "b", "b")],
        ("_p_", "x", "y", "z"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_simple_boolean_query(solver):
    pchoice_as_sets = {Z: {(0.6, "s1"), (0.4, "s2")}}
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    query = Implication(ans(), Z(x))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set([(1.0,)], ("_p_",))
    assert testing.eq_prov_relations(result, expected)


def test_boolean_query_equality(solver):
    pfacts_as_sets = {Z: {(0.6, "s1"), (0.4, "s2")}}
    cpl_program = CPLogicProgram()
    for pred_symb, pfacts_as_set in pfacts_as_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfacts_as_set
        )
    query = Implication(ans(), Z(Constant[str]("s1")))
    result = solver.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set([(.6,)], ("_p_",))
    assert testing.eq_prov_relations(result, expected)


def test_dalvi_suciu_fails_unate():
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.2, 'a'), (0.1, 'b')]
    )
    cpl.add_probabilistic_facts_from_tuples(
        R,
        [(0.1, 'a'), (0.3, 'c')]
    )
    rule = Implication(P(x), Conjunction((R(y), Q(x), Negation(R(x)))))
    cpl.walk(rule)
    query = Implication(ans(x), P(x))
    with pytest.raises(NonLiftableException):
        dalvi_suciu_lift.solve_succ_query(query, cpl)


def test_dalvi_suciu_fails_single_quantifier():
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.2, 'a'), (0.1, 'b')]
    )
    cpl.add_probabilistic_facts_from_tuples(
        R,
        [(0.1, 'a'), (0.3, 'c')]
    )
    cpl.add_probabilistic_facts_from_tuples(
        P,
        [(0.1, 'a'), (0.3, 'c')]
    )
    rule = Implication(A(x), Conjunction((Q(y), R(x))))
    cpl.walk(rule)
    query = Implication(ans(), Conjunction((P(x), Negation(A(y)))))
    with pytest.raises(NonLiftableException):
        dalvi_suciu_lift.solve_succ_query(query, cpl)


def test_dalvi_suciu_fails_not_ranked():
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [(0.2, 'a', 'b'), (0.1, 'b', 'c')]
    )
    rule = Implication(P(x), Conjunction((Q(x, y), Q(y, x))))
    cpl.walk(rule)
    query = Implication(ans(x), P(x))
    with pytest.raises(NonLiftableException):
        dalvi_suciu_lift.solve_succ_query(query, cpl)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_nested_negation(solver):
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [
            (0.2, 'a', 1, 2),
            (0.8, 'b', 1, 5),
            (0.1, 'b', 2, 2),
        ]
    )
    cpl.add_probabilistic_facts_from_tuples(
        R,
        [(0.1, 'a'), (0.3, 'c')]
    )
    program = Union((
        Implication(Z(x, y), Q(x, y, z)),
    ))
    query = Implication(ans(x), Conjunction((
            R(x),
            Negation(Z(x, Constant(1)))
        )))
    cpl.walk(program)

    if solver is small_dichotomy_theorem_based_solver:
        context = pytest.raises(NotHierarchicalQueryException)
    else:
        context = nullcontext()

    with context:
        res = solver.solve_succ_query(query, cpl)

        expected = ProvenanceAlgebraSet(
            Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
                ("_p_", "x"),
                [
                    (0.08, "a"),
                    (0.30, "c"),
                ],
            )),
            str2columnstr_constant("_p_"),
        )

        assert testing.eq_prov_relations(res, expected)


@pytest.mark.parametrize("solver", [
    pytest.param(weighted_model_counting, marks=pytest.mark.xfail(
        reason="WMC issue to be resolved"
    )),
    small_dichotomy_theorem_based_solver,
    dalvi_suciu_lift,
])
def test_nested_negated_existential(solver):
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        Q,
        [
            (0.2, 'a', 1, 2),
            (0.8, 'b', 1, 5),
            (0.1, 'b', 2, 2),
        ]
    )
    cpl.add_probabilistic_facts_from_tuples(
        R,
        [(0.1, 'a'), (0.3, 'c')]
    )
    program = Union((
        Implication(Z(x), Q(x, y, z)),
    ))
    query = Implication(ans(x), Conjunction((
            R(x),
            Negation(Z(x))
        )))
    cpl.walk(program)

    if solver is small_dichotomy_theorem_based_solver:
        context = pytest.raises(NotHierarchicalQueryException)
    else:
        context = nullcontext()

    with context:
        res = solver.solve_succ_query(query, cpl)

        expected = ProvenanceAlgebraSet(
            Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
                ("_p_", "x"),
                [
                    (0.08, "a"),
                    (0.30, "c"),
                ],
            )),
            str2columnstr_constant("_p_"),
        )

        assert testing.eq_prov_relations(res, expected)


def test_disjunctive_query_with_probchoice():
    pchoice_as_sets = {
        Q: {(0.6, 1), (0.4, 2)},
        A: {(0.7, 1), (0.2, 2), (0.1, 3)},
    }
    pfact_as_sets = {
        P: {(0.9, 3), (0.8, 2)},
        R: {(0.4, 1), (0.1, 2)},
    }
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_as_set in pfact_as_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfact_as_set
        )
    code = Union((
        Implication(Z(x), Conjunction((P(x), Q(x), A(y)))),
        Implication(Z(x), Conjunction((R(x), Q(x)))),
    ))
    cpl_program.walk(code)
    query = Implication(ans(x), Z(x))
    result = dalvi_suciu_lift.solve_succ_query(query, cpl_program)
    expected = testing.make_prov_set(
        [
            (0.6 * 0.4, 1),
            (0.4 * (1 - (1 - 0.8) * (1 - 0.1)), 2),
        ],
        ("_p_", "x"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_small_dichotomy_fails_on_disjunctive_query():
    solver = small_dichotomy_theorem_based_solver
    pchoice_as_sets = {
        Q: {(0.6, 1), (0.4, 2)},
        A: {(0.7, 1), (0.2, 2), (0.1, 3)},
    }
    pfact_as_sets = {
        P: {(0.9, 3), (0.8, 2)},
        R: {(0.4, 1), (0.1, 2)},
    }
    cpl_program = CPLogicProgram()
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    for pred_symb, pfact_as_set in pfact_as_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfact_as_set
        )
    code = Union((
        Implication(Z(x), Conjunction((P(x), Q(x), A(y)))),
        Implication(Z(x), Conjunction((R(x), Q(x)))),
    ))
    cpl_program.walk(code)
    query = Implication(ans(x), Z(x))
    with pytest.raises(UnsupportedSolverError):
        solver.solve_succ_query(query, cpl_program)


def test_small_dichotomy_fails_on_noisy_or_projection():
    solver = small_dichotomy_theorem_based_solver
    pfact_as_sets = {
        P: {(0.9, 3, 2), (0.8, 3, 1)},
    }
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_as_set in pfact_as_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfact_as_set
        )
    query = Implication(ans(x), P(x, y))
    with pytest.raises(UnsupportedSolverError):
        solver.solve_succ_query(query, cpl_program)


def test_small_dichotomy_fails_on_noisy_or_projection_with_pchoice_in_query():
    solver = small_dichotomy_theorem_based_solver
    pfact_as_sets = {
        P: {(0.9, 3, 2), (0.8, 3, 1)},
    }
    pchoice_as_sets = {
        Q: {(0.6, 1), (0.4, 2)},
    }
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_as_set in pfact_as_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfact_as_set
        )
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    query = Implication(ans(x), Conjunction((Q(x), P(x, y))))
    with pytest.raises(UnsupportedSolverError):
        solver.solve_succ_query(query, cpl_program)


def test_dalvi_suciu_fails_builtin():
    solver = dalvi_suciu_lift
    pfact_as_sets = {
        P: {(0.9, 3, 2), (0.8, 3, 1)},
    }
    pchoice_as_sets = {
        Q: {(0.6, 1), (0.4, 2)},
    }
    cpl_program = CPLogicProgram()
    for pred_symb, pfact_as_set in pfact_as_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(
            pred_symb, pfact_as_set
        )
    for pred_symb, pchoice_as_set in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(
            pred_symb, pchoice_as_set
        )
    query = Implication(ans(x), Conjunction((Q(x), P(x, y), GT(x, y))))
    with pytest.raises(UnsupportedSolverError):
        solver.solve_succ_query(query, cpl_program)


def test_deterministic_simplification():
    pchoice_as_sets = {
        Q: [(0.6, 'b'), (0.4, 'c')],
    }
    deterministic_as_sets = {
        P: [('a', 'b'), ('b', 'c')]
    }
    cpl_program = CPLogicProgram()
    for k, v in pchoice_as_sets.items():
        cpl_program.add_probabilistic_choice_from_tuples(k, v)
    for k, v in deterministic_as_sets.items():
        cpl_program.add_extensional_predicate_from_tuples(k, v)

    query = Implication(
        ans(x, y),
        Conjunction((P(x, y), P(Constant('a'), y), Q(y)))
    )

    res = dalvi_suciu_lift.solve_succ_query(
        query, cpl_program
    )

    assert testing.eq_prov_relations(
        res, testing.make_prov_set({(0.6, 'a', 'b')}, ['_p_', 'x', 'y'])
    )



import pytest
from .. import config
from ...utils.relational_algebra_set import pandas

@pytest.fixture
def clean_config_fixture():
    backend = config["RAS"]["backend"]
    try:
        yield
    finally:
        config.set_query_backend(backend)

def test_config_set_query_backend(clean_config_fixture):
    with pytest.raises(ValueError):
        config.set_query_backend("hello")

    # config.set_query_backend("dask")
    # assert config["RAS"]["backend"] == "dask"
    # from ...utils.relational_algebra_set import RelationalAlgebraFrozenSet
    # assert RelationalAlgebraFrozenSet is dask_sql.RelationalAlgebraFrozenSet

    config.set_query_backend("pandas")
    assert config["RAS"]["backend"] == "pandas"
    from ...utils.relational_algebra_set import RelationalAlgebraFrozenSet
    assert RelationalAlgebraFrozenSet is pandas.RelationalAlgebraFrozenSet


def test_switch_expression_type_printing():
    old = config.expression_type_printing()
    config.switch_expression_type_printing()
    assert (
        config.expression_type_printing()
        is not old
    )

    config.switch_expression_type_printing()
    assert (
        config.expression_type_printing()
        is old
    )

    config.enable_expression_type_printing()
    assert config.expression_type_printing()

    config.disable_expression_type_printing()
    assert not config.expression_type_printing()



import operator
from typing import AbstractSet

from pytest import fixture

from ...datalog.translate_to_named_ra import TranslateToNamedRA
from ...datalog.wrapped_collections import WrappedRelationalAlgebraSet
from ...expressions import Constant, Symbol
from ...logic import Conjunction, Negation, Disjunction
from ...relational_algebra.relational_algebra import RelationalAlgebraSolver
from ..ranking import verify_that_the_query_is_ranked, partially_rank_query

EQ = Constant(operator.eq)

P = Symbol("P")
Q = Symbol("Q")
R = Symbol("R")
x = Symbol("x")
y = Symbol("y")
z = Symbol("z")


def test_query_ranked():
    query = Conjunction(
        (Q(x, y), Q(x, z), Q(y, z), P(x, y), Negation(Q(y, z)))
    )
    assert verify_that_the_query_is_ranked(query)

    query = Disjunction((Q(x, y), Conjunction((Q(x, z), P(z, y)))))

    assert verify_that_the_query_is_ranked(query)


def test_query_not_ranked():
    query = Conjunction(
        (Q(x, y), Q(x, z), Q(z, x), P(x, y), Negation(Q(y, z)))
    )
    assert not verify_that_the_query_is_ranked(query)

    query = Conjunction(
        (Q(x, y), Q(x, z), P(z, x), P(x, y), Negation(Q(y, z)))
    )
    assert not verify_that_the_query_is_ranked(query)

    query = Disjunction((Q(x, y), Conjunction((Q(x, z), P(y, y)))))

    assert not verify_that_the_query_is_ranked(query)

    query = Conjunction((Q(x, z), P(y, x), EQ(x, y)))

    assert not verify_that_the_query_is_ranked(query)


@fixture
def R1():
    return Constant[AbstractSet](
        WrappedRelationalAlgebraSet([(chr(ord('a') + i),) for i in range(10)])
    )


@fixture
def R2():
    return Constant[AbstractSet](
        WrappedRelationalAlgebraSet([
            (i, chr(ord('a') + i * 2), i * 2)
            for i in range(10)
        ])
    )


@fixture
def R3():
    return Constant[AbstractSet](
        WrappedRelationalAlgebraSet([
            (i,)
            for i in range(10)
        ])
    )


@fixture
def R4():
    return Constant[AbstractSet](
        WrappedRelationalAlgebraSet([
            (
                chr(ord('a') + i * 2),
                chr(ord('a') + i)
            )
            for i in range(10)
        ])
    )


def test_ranking_repeated(R2, R4):
    query = Conjunction((R(x, y, z), Q(y, y)))
    ra_query = TranslateToNamedRA().walk(query)

    symbol_table = dict({Q: R4, R: R2})
    ras = RelationalAlgebraSolver(symbol_table)
    ra_sol = ras.walk(ra_query)

    res = partially_rank_query(query, symbol_table)

    assert verify_that_the_query_is_ranked(res)

    ras = RelationalAlgebraSolver(symbol_table)
    ra_query = TranslateToNamedRA().walk(res)
    ranked_sol = ras.walk(ra_query)

    assert ra_sol == ranked_sol

    query = Conjunction((R(x, y, x), Q(z, y)))
    ra_query = TranslateToNamedRA().walk(query)

    symbol_table = dict({Q: R4, R: R2})
    ras = RelationalAlgebraSolver(symbol_table)
    ra_sol = ras.walk(ra_query)

    res = partially_rank_query(query, symbol_table)

    assert verify_that_the_query_is_ranked(res)

    ras = RelationalAlgebraSolver(symbol_table)
    ra_query = TranslateToNamedRA().walk(res)
    ranked_sol = ras.walk(ra_query)

    assert ra_sol == ranked_sol


def test_cant_rank_repeated():
    query = Conjunction((Q(x, z), Q(y, y)))
    symbol_table = dict({Q: R4, R: R2})

    res = partially_rank_query(query, symbol_table)

    assert not verify_that_the_query_is_ranked(res)



"""
Compiler for the intermediate representation of a Datalog program.
The :class:`DatalogProgram` class processes the
intermediate representation of a program and extracts
the extensional, intensional, and builtin
sets and has support for constraints.
"""

from typing import Iterable

from .expression_processing import extract_logic_atoms
from .basic_representation import DatalogProgramMixin
from ..expression_walker import ExpressionWalker, PatternWalker, add_match
from ..expressions import Constant, Symbol
from ..logic import LogicOperator, Union

class RightImplication(LogicOperator):
    """
    This class defines implications to the right. They are used to define
    constraints in datalog programs. The functionality is the same as
    that of an implication, but with body and head inverted in position.
    """

    def __init__(self, antecedent, consequent):
        self.antecedent = antecedent
        self.consequent = consequent
        self._symbols = consequent._symbols | antecedent._symbols

    def __repr__(self):
        return "RightImplication{{{} \u2192 {}}}".format(
            repr(self.antecedent), repr(self.consequent)
        )


class DatalogConstraintsMixin(PatternWalker):
    protected_keywords = {"__constraints__"}
    categorized_constraints = {}
    existential_rules = {}

    @add_match(RightImplication)
    def add_logic_constraint(self, expression):
        sym_constraints = Symbol("__constraints__")
        if (
            isinstance(expression, RightImplication)
            and sym_constraints in self.symbol_table
        ):
            constrains = self.symbol_table[sym_constraints]
            constrains = Union((constrains.formulas + (expression,)))
            self.symbol_table[sym_constraints] = constrains
        elif isinstance(expression, RightImplication):
            self.symbol_table[sym_constraints] = Union((expression,))

        return expression

    def constraints(self):
        '''Function that returns the constraints contained
        in the Datalog program.

        Returns
        -------
        Union of formulas containing all constraints loaded into the program.
        '''
        sym_constraints = Symbol("__constraints__")
        return self.symbol_table.get(sym_constraints, Union(()))

    def set_constraints(self, categorized_constraints):
        '''This function receives a dictionary with the contraints organized
        according to the functor of its consequent and is in charge of setting
        it both in the symbol table and in the global variable
        `categorized_contraints`, useful for XRewriter optimizations.

        Parameters
        ----------
        categorized_constraints : dict
            Dictionary of constraints where each key is
            the functor of the consequent of the rules
            and the values are lists of contraints with
            each rule associated to the corresponding functor.
        '''
        sym_constraints = Symbol("__constraints__")
        if isinstance(categorized_constraints, dict):
            cons = [b for a in list(categorized_constraints.values()) for b in a]
            self.symbol_table[sym_constraints] = Union(cons)

            self.categorized_constraints = categorized_constraints

    def get_constraints(self):
        '''Returns the contraints in a dictionary, where the key is the functor
        of the consequent of each of the rules.

        Returns
        -------
        Dictionary containing all constraints loaded in the datalog program,
        indexed according to the functor of the rule consequent.
        '''
        sym_constraints = Symbol("__constraints__")
        if not self.categorized_constraints:
            constraints = self.symbol_table.get(sym_constraints, Union(()))
            for f in constraints.formulas:
                self._categorize_constraints(f)

        return self.categorized_constraints

    def add_existential_rules(self, rules):
        self.existential_rules = rules

    def _categorize_constraints(self, sigma):
        '''Function in charge of sorting the constraints in a dictionary
        using the consequent functor as an index.

        This indexing is useful to obtain the constraints in the
        way they are needed for the rewriting algorithm.

        Parameters
        ----------
        sigma : RightImplication
            constraint to be categorized.
        '''
        sigma_functor = sigma.consequent.functor.name
        if (
            self.categorized_constraints
            and sigma_functor in self.categorized_constraints
        ):
            cons_set = self.categorized_constraints[sigma_functor]
            if sigma not in cons_set:
                cons_set.add(sigma)
                self.categorized_constraints[sigma_functor] = cons_set
        else:
            self.categorized_constraints[sigma_functor] = set([sigma])

def reachable_code(query, datalog):
    """Produces the code reachable by a query

    Parameters
    ----------
    query : Implication
        Rule to figure out the reachable program from
    datalog : DatalogProgram
        datalog instance containing the EDB and IDB.

    Returns
    -------
    ExpressionBlock
        Code needed to solve the query.
    """
    if not isinstance(query, Iterable):
        query = [query]

    reachable_code = []
    idb = datalog.intensional_database()
    to_reach = [q.consequent.functor for q in query]
    reached = set()
    seen_rules = set()
    while to_reach:
        p = to_reach.pop()
        reached.add(p)
        rules = idb[p]
        if isinstance(rules, Constant):
            continue
        for rule in rules.formulas:
            if (
                rule in seen_rules or
                isinstance(rule, RightImplication)
            ):
                continue
            seen_rules.add(rule)
            reachable_code.append(rule)
            for predicate in extract_logic_atoms(rule.antecedent):
                functor = predicate.functor
                if functor not in reached and functor in idb:
                    to_reach.append(functor)

    return Union(reachable_code[::-1])

class DatalogConstraintsProgram(
    DatalogConstraintsMixin,
    DatalogProgramMixin,
    ExpressionWalker
):
    pass


from collections.abc import MutableSet, Sequence
from itertools import tee


class OrderedSet(MutableSet, Sequence):
    def __init__(self, iterable=None):
        if iterable is None:
            iterable = []

        it1, it2 = tee(iterable)
        self._set = set(it1)
        self._list = []
        for i in it2:
            if i not in self._list:
                self._list.append(i)

    def add(self, value):
        if value not in self._set:
            self._set.add(value)
            self._list.append(value)

    def discard(self, value):
        if value in self._set:
            self._set.discard(value)
            self._list.remove(value)

    def replace(self, src, dst):
        src_ix = self._list.index(src)
        self._list = (
            self._list[:src_ix] +
            [dst] +
            self._list[src_ix + 1:]
        )
        self._set.discard(src)
        self._set.add(dst)

    def issubset(self, other):
        return self <= other

    def issuperset(self, other):
        return self >= other

    def __contains__(self, value):
        return value in self._set

    def __len__(self):
        return len(self._set)

    def __iter__(self):
        return iter(self._list)

    def __getitem__(self, slice_):
        return self._list[slice_]

    def index(self, value, start=0, stop=None):
        return self._list[start: stop].index(value)

    def __repr__(self):
        return 'OrderedSet(' + ','.join(repr(i) for i in self) + ')'

    def copy(self):
        res = OrderedSet()
        res._set = self._set.copy()
        res._list = self._list.copy()
        return res



from ....datalog.expression_processing import extract_logic_atoms
from ....expression_walker import IdentityWalker, PatternWalker, add_match
from ....expressions import Symbol
from ....logic import Implication

RESAMPLE = Symbol("RESAMPLE")


class TranslateResamplingMixin(PatternWalker):
    @add_match(
        Implication,
        lambda e: any(
            a.functor is RESAMPLE for a in extract_logic_atoms(e.antecedent)
        )
    )
    def translate_resampling(self, expression):
        resample_set = 



from pathlib import Path
from typing import Optional

import numpy as np
import pandas as pd
from nilearn.datasets.utils import _fetch_files
from scipy import sparse


class StudyID(str):
    pass


class TfIDf(float):
    pass


NS_DATA_URL = "https://github.com/neurosynth/neurosynth-data/raw/master/"


def fetch_study_metadata(
    data_dir: Path, version: int = 7, verbose: int = 1
) -> pd.DataFrame:
    """
    Download if needed the `metadata.tsv.gz` file from Neurosynth and load
    it into a pandas DataFrame.

    The metadata table contains the metadata for each study. Each study (ID)
    is stored on its own line. These IDs are in the same order as the id
    column of the associated `coordinates.tsv.gz` file, but the rows will
    differ because the coordinates file will contain multiple rows per
    study. They are also in the same order as the rows in the
    `features.npz` files for the same version.

    The metadata will therefore have N rows, N being the number of studies
    in the Neurosynth dataset. The columns (for version 7) are:
        - id
        - doi
        - space
        - title
        - authors
        - year
        - journal

    Parameters
    ----------
    data_dir : Path
        the path for the directory where downloaded data should be saved.
    version : int, optional
        the neurosynth data version, by default 7
    verbose : int, optional
        verbose param for nilearn's `_fetch_files`, by default 1

    Returns
    -------
    pd.DataFrame
        the study metadata dataframe
    """
    metadata_filename = f"data-neurosynth_version-{version}_metadata.tsv.gz"
    metadata_file = _fetch_files(
        data_dir,
        [
            (
                metadata_filename,
                NS_DATA_URL + metadata_filename,
                {},
            ),
        ],
        verbose=verbose,
    )[0]
    metadata = pd.read_table(metadata_file)
    return metadata


def fetch_feature_data(
    data_dir: Path,
    version: int = 7,
    verbose: int = 1,
    convert_study_ids: bool = False,
) -> pd.DataFrame:
    """
    Download if needed the `tfidf_features.npz` file from Neurosynth and
    load it into a pandas Dataframe. The `tfidf_features` contains feature
    values for different types of "vocabularies".

    The features dataframe is stored as a compressed, sparse matrix.
    Once loaded and reconstructed into a dense matrix, it contains one row
    per study and one column per label. The associated labels are loaded,
    as well as the study ids, to reconstruct a dataframe of size N x P,
    where N is the number of studies in the Neurosynth dataset, and P is
    the number of words in the vocabulary.

    Parameters
    ----------
    data_dir : Path
        the path for the directory where downloaded data should be saved.
    version : int, optional
        the neurosynth data version, by default 7
    verbose : int, optional
        verbose param for nilearn's `_fetch_files`, by default 1
    convert_study_ids : bool, optional
        if True, cast study ids as `StudyID`, by default False

    Returns
    -------
    pd.DataFrame
        the features dataframe
    """
    file_names = [
        f"data-neurosynth_version-{version}_vocab-terms_source-abstract_type-tfidf_features.npz",
        f"data-neurosynth_version-{version}_vocab-terms_vocabulary.txt",
    ]
    files = _fetch_files(
        data_dir,
        [
            (
                fn,
                NS_DATA_URL + fn,
                {},
            )
            for fn in file_names
        ],
        verbose=verbose,
    )
    feature_data_sparse = sparse.load_npz(files[0])
    feature_data = feature_data_sparse.todense()
    metadata_df = fetch_study_metadata(data_dir, version, verbose)
    ids = metadata_df["id"]
    if convert_study_ids:
        ids = ids.apply(StudyID)
    feature_names = np.genfromtxt(
        files[1],
        dtype=str,
        delimiter="\t",
    ).tolist()
    feature_df = pd.DataFrame(
        index=ids.tolist(), columns=feature_names, data=feature_data
    )
    return feature_df


def fetch_neurosynth_peak_data(
    data_dir: Path,
    version: int = 7,
    verbose: int = 1,
    convert_study_ids: bool = False,
) -> pd.DataFrame:
    """
    Download if needed the `coordinates.tsv.gz` file from Neurosynth and
    load it into a pandas DataFrame.

    The `coordinates.tsv.gz` contains the coordinates for the peaks
    reported by studies in the Neurosynth dataset. It contains one row per
    coordinate reported.

    The metadata for each study is also loaded to include the space in
    which the coordinates are reported. The peak_data dataframe therefore
    has PR rows, PR being the number of reported peaks in the Neurosynth
    dataset.

    The columns (for version 7) are:
        - id
        - table_id
        - table_num
        - peak_id
        - space
        - x
        - y
        - z

    Parameters
    ----------
    data_dir : Path
        the path for the directory where downloaded data should be saved.
    version : int, optional
        the neurosynth data version, by default 7
    verbose : int, optional
        verbose param for nilearn's `_fetch_files`, by default 1
    convert_study_ids : bool, optional
        if True, cast study ids as `StudyID`, by default False

    Returns
    -------
    pd.DataFrame
        the peak dataframe
    """
    coordinates_filename = (
        f"data-neurosynth_version-{version}_coordinates.tsv.gz"
    )
    coordinates_file = _fetch_files(
        data_dir,
        [
            (
                coordinates_filename,
                NS_DATA_URL + coordinates_filename,
                {},
            ),
        ],
        verbose=verbose,
    )[0]
    activations = pd.read_table(coordinates_file)
    metadata = fetch_study_metadata(data_dir, version, verbose)
    activations = activations.join(
        metadata[["id", "space"]].set_index("id"), on="id"
    )
    if convert_study_ids:
        activations["id"] = activations["id"].apply(StudyID)
    return activations


def get_ns_term_study_associations(
    data_dir: Path,
    version: int = 7,
    verbose: int = 1,
    convert_study_ids: bool = False,
    tfidf_threshold: Optional[float] = None,
) -> pd.DataFrame:
    """
    Load a dataframe containing associations between term and studies.
    The dataframe contains one row for each term and study pair from the
    features table in the Neurosynth dataset. With each (term, study) pair
    comes the tfidf value for the term in the study.
    If a tfidf threshold value is passed, only (term, study) associations
    with a tfidf value > tfidf_threshold will be kept.

    Parameters
    ----------
    data_dir : Path
        the path for the directory where downloaded data should be saved.
    version : int, optional
        the neurosynth data version, by default 7
    verbose : int, optional
        verbose param for nilearn's `_fetch_files`, by default 1
    convert_study_ids : bool, optional
        if True, cast study ids as `StudyID`, by default False
    tfidf_threshold : Optional[float], optional
        the minimum tfidf value for the (term, study) associations,
        by default None

    Returns
    -------
    pd.DataFrame
        the term association dataframe
    """
    features = fetch_feature_data(
        data_dir, version, verbose, convert_study_ids
    )
    features.index.name = "id"
    term_data = pd.melt(
        features.reset_index(),
        var_name="term",
        id_vars="id",
        value_name="tfidf",
    )
    if tfidf_threshold is not None:
        term_data = term_data.query(f"tfidf > {tfidf_threshold}")
    else:
        term_data = term_data.query("tfidf > 0")
    return term_data


def get_ns_mni_peaks_reported(
    data_dir: Path,
    version: int = 7,
    verbose: int = 1,
    convert_study_ids: bool = False,
) -> pd.DataFrame:
    """
    Load a dataframe containing the coordinates for the peaks reported by
    studies in the Neurosynth dataset. Coordinates for the peaks are in
    MNI space, with coordinates that are reported in Talaraich space
    converted.

    The resulting dataframe contains one row for each peak reported. Each
    row has 4 columns:
        - id
        - x
        - y
        - z

    Parameters
    ----------
    data_dir : Path
        the path for the directory where downloaded data should be saved.
    version : int, optional
        the neurosynth data version, by default 7
    verbose : int, optional
        verbose param for nilearn's `_fetch_files`, by default 1
    convert_study_ids : bool, optional
        if True, cast study ids as `StudyID`, by default False

    Returns
    -------
    pd.DataFrame
        the peak dataframe
    """
    activations = fetch_neurosynth_peak_data(
        data_dir, version, verbose, convert_study_ids
    )
    mni_peaks = activations.loc[activations.space == "MNI"][
        ["x", "y", "z", "id"]
    ]
    non_mni_peaks = activations.loc[activations.space == "TAL"][
        ["x", "y", "z", "id"]
    ]
    proj_mat = np.linalg.pinv(
        np.array(
            [
                [0.9254, 0.0024, -0.0118, -1.0207],
                [-0.0048, 0.9316, -0.0871, -1.7667],
                [0.0152, 0.0883, 0.8924, 4.0926],
                [0.0, 0.0, 0.0, 1.0],
            ]
        ).T
    )
    projected = np.round(
        np.dot(
            np.hstack(
                (
                    non_mni_peaks[["x", "y", "z"]].values,
                    np.ones((len(non_mni_peaks), 1)),
                )
            ),
            proj_mat,
        )[:, 0:3]
    )
    projected_df = pd.DataFrame(
        np.hstack([projected, non_mni_peaks[["id"]].values]),
        columns=["x", "y", "z", "id"],
    )
    peak_data = pd.concat([projected_df, mni_peaks]).astype(
        {"x": int, "y": int, "z": int}
    )
    return peak_data
